#=========================================================================
# ImportPass.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 25, 2019
"""Provide a pass that imports arbitrary SystemVerilog modules."""

import copy
import importlib
import linecache
import os
import shutil
import subprocess
import sys
from textwrap import indent

from pymtl3.datatypes import Bits, is_bitstruct_class, is_bitstruct_inst, mk_bits
from pymtl3.dsl import Component
from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir import get_component_ifc_rtlir

from ..errors import SVerilogImportError
from ..util.utility import expand, get_component_unique_name, make_indent, wrap
from .ImportConfigs import ImportConfigs

try:
  # Python 2
  reload
except NameError:
  # Python 3
  from importlib import reload

class ImportPass( BasePass ):
  """Import an arbitrary SystemVerilog module as a PyMTL component.

  The import pass takes as input a PyMTL component hierarchy where
  the components to be imported have an `import` entry in their parameter
  dictionary( set by calling the `set_parameter` PyMTL API ).
  This pass assumes the modules to be imported are located in the current
  directory and have name {full_name}.sv where `full_name` is the name of
  component's class concatanated with the list of arguments of its construct
  method. It has the following format:
      > {module_name}[__{parameter_name}_{parameter_value}]*
  As an example, component mux created through `mux = Mux(Bits32, 2)` has a
  full name `Mux__Type_Bits32__ninputs_2`.
  The top module inside the target .sv file should also have a full name.
  """
  def __call__( s, top ):
    s.top = top
    if not top._dsl.constructed:
      raise SVerilogImportError( top,
        f"please elaborate design {top} before applying the import pass!" )
    ret = s.traverse_hierarchy( top )
    if ret is None:
      ret = top
    return ret

  def traverse_hierarchy( s, m ):
    if hasattr(m, f"config_{s.get_backend_name()}_import") and \
       isinstance(s.get_config(m), ImportConfigs) and \
       s.get_config(m).import_:
      s.get_config(m).fill_missing( m )
      s.get_config(m).check()
      return s.do_import( m )
    else:
      for child in m.get_child_components():
        s.traverse_hierarchy( child )

  def do_import( s, m ):
    try:
      imp = s.get_imported_object( m )
      if m is s.top:
        return imp
      else:
        s.top.replace_component_with_obj( m, imp )
    except AssertionError as e:
      msg = '' if e.args[0] is None else e.args[0]
      raise SVerilogImportError( m, msg )

  #-----------------------------------------------------------------------
  # Backend-specific methods
  #-----------------------------------------------------------------------

  def get_backend_name( s ):
    return "sverilog"

  def get_config( s, m ):
    return m.config_sverilog_import

  def get_translation_namespace( s, m ):
    return m._pass_sverilog_translation

  #-----------------------------------------------------------------------
  # is_cached
  #-----------------------------------------------------------------------

  def is_cached( s, m, full_name ):
    cached = False

    # Only components translated by pymtl translation passes will be cached
    try:
      is_same = s.get_translation_namespace(m).is_same
    except AttributeError:
      is_same = False

    # Check if the verilated model is cached
    cached = False
    obj_dir = 'obj_dir_' + full_name
    c_wrapper = full_name + '_v.cpp'
    py_wrapper = full_name + '_v.py'
    shared_lib = f'lib{full_name}_v.so'
    if is_same and os.path.exists(obj_dir) and os.path.exists(c_wrapper) and \
       os.path.exists(py_wrapper) and os.path.exists(shared_lib):
      cached = True

    return cached

  #-----------------------------------------------------------------------
  # get_imported_object
  #-----------------------------------------------------------------------

  def get_imported_object( s, m ):
    config = s.get_config( m )
    rtype = get_component_ifc_rtlir( m )
    full_name = get_component_unique_name( rtype )
    p_map = config.get_port_map() if config.is_port_mapped() else lambda x: x
    no_clk, no_reset = not config.has_clk, not config.has_reset
    _packed_ports = s.gen_packed_ports( rtype )
    clk = next(filter(lambda x: x[0]=='clk', _packed_ports))[1]
    reset = next(filter(lambda x: x[0]=='reset', _packed_ports))[1]
    packed_ports = \
        ([('clk', '' if no_clk else p_map('clk'), clk)] if no_clk else []) + \
        ([('reset', '' if no_reset else p_map('reset'), reset)] if no_reset else []) + \
        [ (n, p_map(n), p) for n, p in _packed_ports \
          if not (n == 'clk' and no_clk or n == 'reset' and no_reset)]

    cached = s.is_cached( m, full_name )

    # Create a new Verilog source file if a new top-level wrapper is needed
    if config.is_top_wrapper():
      s.add_param_wrapper( m, config, rtype, packed_ports )

    s.create_verilator_model( m, config, cached )

    port_cdefs = \
        s.create_verilator_c_wrapper( m, config, packed_ports, cached )

    s.create_shared_lib( m, config, cached )

    symbols = \
        s.create_py_wrapper( m, config, rtype, packed_ports, port_cdefs, cached )

    imp = s.import_component( m, config, symbols )

    return imp

  #-----------------------------------------------------------------------
  # create_verilator_model
  #-----------------------------------------------------------------------

  def create_verilator_model( s, m, config, cached ):
    """Verilate module `m`."""
    config.vprint("\n=====Verilate model=====")
    if not cached:
      # Generate verilator command
      cmd = config.create_vl_cmd()

      # Remove obj_dir directory if it already exists.
      # obj_dir is where the verilator output ( C headers and sources ) is stored
      obj_dir = config.vl_mk_dir
      if os.path.exists( obj_dir ):
        shutil.rmtree( obj_dir )

      # Try to call verilator
      try:
        config.vprint(f"Verilating {config.top_module} with command:", 2)
        config.vprint(f"{cmd}", 4)
        subprocess.check_output( cmd, stderr = subprocess.STDOUT, shell = True )
      except subprocess.CalledProcessError as e:
        err_msg = e.output if not isinstance(e.output, bytes) else \
                  e.output.decode('utf-8')
        import_err_msg = \
            f"Fail to verilate model {config.top_module}\n"\
            f"  Verilator command:\n{indent(cmd, '  ')}\n\n"\
            f"  Verilator output:\n{indent(wrap(err_msg), '  ')}\n"
        raise SVerilogImportError(m, import_err_msg) from e
      config.vprint(f"Successfully verilated the given model!", 2)

    else:
      config.vprint(f"{config.top_module} not verilated because it's cached!", 2)

  #-----------------------------------------------------------------------
  # create_verilator_c_wrapper
  #-----------------------------------------------------------------------

  def create_verilator_c_wrapper( s, m, config, packed_ports, cached ):
    """Return the file name of generated C component wrapper.

    Create a C wrapper that calls verilator C API and provides interfaces
    that can be later called through CFFI.
    """
    component_name = config.top_module
    dump_vcd = int(config.vl_trace)
    vcd_timescale = config.vl_trace_timescale
    half_cycle_time = config.vl_trace_cycle_time // 2
    external_trace = int(config.external_trace)
    wrapper_name = config.get_c_wrapper_path()
    verilator_xinit_value = config.get_vl_xinit_value()
    config.vprint("\n=====Generate C wrapper=====")

    # The wrapper template should be in the same directory as this file
    template_name = \
      os.path.dirname( os.path.abspath( __file__ ) ) + \
      os.path.sep + 'verilator_wrapper.c.template'

    # Generate port declarations for the verilated model in C
    port_defs = []
    for name, v_name, port in packed_ports:
      if v_name:
        port_defs.append( s.gen_signal_decl_c( v_name, port ) )
    port_cdefs = copy.copy( port_defs )
    make_indent( port_defs, 2 )
    port_defs = '\n'.join( port_defs )

    # Generate initialization statements for in/out ports
    port_inits = []
    for name, v_name, port in packed_ports:
      if v_name:
        port_inits.extend( s.gen_signal_init_c( v_name, port ) )
    make_indent( port_inits, 1 )
    port_inits = '\n'.join( port_inits )

    # Fill in the C wrapper template

    # Since we may run import with or without dump_vcd enabled, we need
    # to dump C wrapper regardless of whether the verilated model is
    # cached or not.
    # TODO: we can avoid dumping C wrapper if we attach some metadata to
    # tell if the wrapper was generated with or without `dump_vcd` enabled.
    with open( template_name, 'r' ) as template:
      with open( wrapper_name, 'w' ) as output:
        c_wrapper = template.read()
        c_wrapper = c_wrapper.format( **locals() )
        output.write( c_wrapper )

    config.vprint(f"Successfully generated C wrapper {wrapper_name}!", 2)
    return port_cdefs

  #-----------------------------------------------------------------------
  # create_shared_lib
  #-----------------------------------------------------------------------

  def create_shared_lib( s, m, config, cached ):
    """Return the name of compiled shared lib."""
    full_name = config.top_module
    dump_vcd = config.vl_trace
    config.vprint("\n=====Compile shared library=====")

    # Since we may run import with or without dump_vcd enabled, we need
    # to compile C wrapper regardless of whether the verilated model is
    # cached or not.
    # TODO: A better caching strategy is to attach some metadata
    # to the C wrapper so that we know the wrapper was generated with or
    # without dump_vcd enabled.
    if dump_vcd or not cached:
      cmd = config.create_cc_cmd()

      # Try to call the C compiler
      try:
        config.vprint("Compiling shared library with command:", 2)
        config.vprint(f"{cmd}", 4)
        subprocess.check_output( cmd, stderr = subprocess.STDOUT, shell = True,
                                 universal_newlines=True )
      except subprocess.CalledProcessError as e:
        err_msg = e.output if not isinstance(e.output, bytes) else \
                  e.output.decode('utf-8')
        import_err_msg = \
            f"Failed to compile Verilated model into a shared library:\n"\
            f"  C compiler command:\n{indent(cmd, '  ')}\n\n"\
            f"  C compiler output:\n{indent(wrap(err_msg), '  ')}\n"
        raise SVerilogImportError(m, import_err_msg) from e
      config.vprint(f"Successfully compiled shared library "\
                    f"{config.get_shared_lib_path()}!", 2)

    else:
      config.vprint(f"Didn't compile shared library because it's cached!", 2)

  #-----------------------------------------------------------------------
  # create_py_wrapper
  #-----------------------------------------------------------------------

  def create_py_wrapper( s, m, config, rtype, packed_ports, port_cdefs, cached ):
    """Return the file name of the generated PyMTL component wrapper."""
    config.vprint("\n=====Generate PyMTL wrapper=====")

    # Load the wrapper template
    template_name = \
      os.path.dirname( os.path.abspath( __file__ ) ) + \
      os.path.sep + 'verilator_wrapper.py.template'
    wrapper_name = config.get_py_wrapper_path()

    # Port definitions of verilated model
    make_indent( port_cdefs, 4 )

    # Port definition in PyMTL style
    symbols, port_defs, connections = s.gen_signal_decl_py( rtype )
    make_indent( port_defs, 2 )
    make_indent( connections, 2 )
    # Wire definition in PyMTL style
    wire_defs = []
    for name, v_name, port in packed_ports:
      wire_defs.append( s.gen_wire_decl_py( name, port ) )
    make_indent( wire_defs, 2 )

    # Set upblk inputs and outputs
    set_comb_input = s.gen_comb_input( packed_ports )
    set_comb_output = s.gen_comb_output( packed_ports )
    make_indent( set_comb_input, 3 )
    make_indent( set_comb_output, 3 )

    # Generate constraints for sequential block
    constraints = s.gen_constraints( packed_ports )
    make_indent( constraints, 4 )
    constraint_str = '' if not constraints else \
"""\
constraint_list = [
{}
      ]

      s.add_constraints( *constraint_list )
""".format( '\n'.join( constraints ) )

    # Line trace
    line_trace = s.gen_line_trace_py( packed_ports )

    # Internal line trace
    in_line_trace = s.gen_internal_line_trace_py( packed_ports )

    # External trace function definition
    if config.external_trace:
      external_trace_c_def = f'void trace( V{config.top_module}_t *, char * );'
    else:
      external_trace_c_def = ''

    # Fill in the python wrapper template
    if not cached:
      with open( template_name, 'r' ) as template:
        with open( wrapper_name, 'w' ) as output:
          py_wrapper = template.read()
          py_wrapper = py_wrapper.format(
            component_name  = config.top_module,
            has_clk         = int(config.has_clk),
            clk             = 'inv_clk' if not config.has_clk else \
                              next(filter(lambda x: x[0]=='clk', packed_ports))[1],
            lib_file        = config.get_shared_lib_path(),
            port_cdefs      = ('  '*4+'\n').join( port_cdefs ),
            port_defs       = '\n'.join( port_defs ),
            wire_defs       = '\n'.join( wire_defs ),
            connections     = '\n'.join( connections ),
            set_comb_input  = '\n'.join( set_comb_input ),
            set_comb_output = '\n'.join( set_comb_output ),
            constraint_str  = constraint_str,
            line_trace      = line_trace,
            in_line_trace   = in_line_trace,
            dump_vcd        = int(config.vl_trace),
            external_trace  = int(config.external_trace),
            trace_c_def     = external_trace_c_def,
          )
          output.write( py_wrapper )

    config.vprint(f"Successfully generated PyMTL wrapper {wrapper_name}!", 2)
    return symbols

  #-----------------------------------------------------------------------
  # import_component
  #-----------------------------------------------------------------------

  def import_component( s, m, config, symbols ):
    """Return the PyMTL component imported from `wrapper_name`.sv."""
    config.vprint("=====Create python object=====")

    component_name = config.top_module
    # Get the name of the wrapper Python module
    wrapper_name = config.get_py_wrapper_path()
    wrapper = wrapper_name.split('.')[0]

    # Add CWD to sys.path so we can import from the current directory
    if not os.getcwd() in sys.path:
      sys.path.append( os.getcwd() )

    # Check linecache in case the wrapper file has been modified
    linecache.checkcache()

    # Import the component from python wrapper

    if wrapper in sys.modules:
      # Reload the wrapper module in case the user has updated the wrapper
      reload(sys.modules[wrapper])
    else:
      # importlib.import_module inserts the wrapper module into sys.modules
      importlib.import_module(wrapper)

    # Try to access the top component class from the wrapper module
    try:
      imp_class = getattr( sys.modules[wrapper], component_name )
    except AttributeError as e:
      raise SVerilogImportError(m,
          f"internal error: PyMTL wrapper {wrapper_name} does not have "
          f"top component {component_name}!") from e

    imp = imp_class()
    config.vprint(f"Successfully created python object of {component_name}!", 2)

    # Update the global namespace of `construct` so that the struct and interface
    # classes defined previously can still be used in the imported model.
    imp.construct.__globals__.update( symbols )

    config.vprint("Import succeeds!")
    return imp

  #-------------------------------------------------------------------------
  # add_param_wrapper
  #-------------------------------------------------------------------------

  def add_param_wrapper( s, m, config, rtype, packed_ports ):
    outfile = f"{config.top_module}.sv"
    parameters = config.v_param

    with open(outfile, "w") as top_wrapper:
      # Port definitions of top-level wrapper
      ports = [
        f"  {p.get_direction()} logic [{p.get_dtype().get_length()}-1:0]"\
        f" {name}{'' if idx == len(packed_ports)-1 else ','}" \
        for idx, (_, name, p) in enumerate(packed_ports) if name
      ]
      # Parameters passed to the module to be parametrized
      params = [
        f"    .{param}( {val} ){'' if idx == len(parameters)-1 else ','}"\
        for idx, (param, val) in enumerate(parameters)
      ]
      # Connections between top module and inner module
      connect_ports = [
        f"    .{name}( {name} ){'' if idx == len(packed_ports)-1 else ','}"\
        for idx, (_, name, p) in enumerate(packed_ports) if name
      ]
      lines = [
        "// This is a top-level module that wraps a parametrized module",
        "// This file is generated by PyMTL SystemVerilog import pass",
        f'`include "{expand(config.get_param_include())}"',
        f"module {config.top_module}",
        "(",
      ] + ports + [
        ");",
        f"  {config.get_module_to_parametrize()}",
        "  #(",
      ] + params + [
        "  ) wrapped_module",
        "  (",
      ] + connect_ports + [
        "  );",
        "endmodule",
      ]
      top_wrapper.write("\n".join(line for line in lines))
      top_wrapper.close()

  #-------------------------------------------------------------------------
  # gen_packed_ports
  #-------------------------------------------------------------------------

  def mangle_port( s, id_, port, n_dim ):
    if not n_dim:
      return [ ( id_, port ) ]
    else:
      return [ ( id_, rt.Array( n_dim, port ) ) ]

  def mangle_interface( s, id_, ifc, n_dim ):
    if not n_dim:
      ret = []
      all_properties = ifc.get_all_properties_packed()
      for name, rtype in all_properties:
        _n_dim, _rtype = s._get_rtype( rtype )
        if isinstance( _rtype, rt.Port ):
          ret += s.mangle_port( id_+"__"+name, _rtype, _n_dim )
        elif isinstance( _rtype, rt.InterfaceView ):
          ret += s.mangle_interface( id_+"__"+name, _rtype, _n_dim )
        else:
          assert False, f"{name} is not interface(s) or port(s)!"
    else:
      ret = []
      for i in range( n_dim[0] ):
        ret += s.mangle_interface( id_+"__"+str(i), ifc, n_dim[1:] )
    return ret

  def gen_packed_ports( s, rtype ):
    """Return a list of (name, rt.Port ) that has all ports of `rtype`.

    This method performs SystemVerilog backend-specific name mangling and
    returns all ports that appear in the interface of component `rtype`.
    Each tuple contains a port or an array of port that has any data type
    allowed in RTLIRDataType.
    """
    packed_ports = []
    ports = rtype.get_ports_packed()
    ifcs = rtype.get_ifc_views_packed()
    for id_, port in ports:
      p_n_dim, p_rtype = s._get_rtype( port )
      packed_ports += s.mangle_port( id_, p_rtype, p_n_dim )
    for id_, ifc in ifcs:
      i_n_dim, i_rtype = s._get_rtype( ifc )
      packed_ports += s.mangle_interface( id_, i_rtype, i_n_dim )
    return packed_ports

  #-------------------------------------------------------------------------
  # gen_signal_decl_c
  #-------------------------------------------------------------------------

  def gen_signal_decl_c( s, name, port ):
    """Return C variable declaration of `port`."""
    c_dim = s._get_c_dim( port )
    nbits = s._get_c_nbits( port )
    UNSIGNED_8  = 'unsigned char'
    UNSIGNED_16 = 'unsigned short'
    UNSIGNED_32 = 'unsigned int'
    if sys.maxsize > 2**32:
      UNSIGNED_64 = 'unsigned long'
    else:
      UNSIGNED_64 = 'unsigned long long'
    if    nbits <= 8:  data_type = UNSIGNED_8
    elif  nbits <= 16: data_type = UNSIGNED_16
    elif  nbits <= 32: data_type = UNSIGNED_32
    elif  nbits <= 64: data_type = UNSIGNED_64
    else:              data_type = UNSIGNED_32
    name = s._verilator_name( name )
    return f'{data_type} * {name}{c_dim};'

  #-------------------------------------------------------------------------
  # gen_signal_init_c
  #-------------------------------------------------------------------------

  def gen_signal_init_c( s, name, port ):
    """Return C port variable initialization."""
    ret       = []
    c_dim     = s._get_c_dim( port )
    nbits     = s._get_c_nbits( port )
    deference = '&' if nbits <= 64 else ''
    name      = s._verilator_name( name )

    if c_dim:
      n_dim_size = s._get_c_n_dim( port )
      sub = ""
      for_template = \
"""\
for ( int i_{idx} = 0; i_{idx} < {dim_size}; i_{idx}++ )
"""
      assign_template = \
"""\
m->{name}{sub} = {deference}model->{name}{sub};
"""

      for idx, dim_size in enumerate( n_dim_size ):
        ret.append( for_template.format( **locals() ) )
        sub += f"[i_{idx}]"

      ret.append( assign_template.format( **locals() ) )

      # Indent the for loop
      for start, dim_size in enumerate( n_dim_size ):
        for idx in range( start + 1, len( n_dim_size ) + 1 ):
          ret[ idx ] = "  " + ret[ idx ]

    else:
      ret.append(f'm->{name} = {deference}model->{name};')

    return ret

  #-----------------------------------------------------------------------
  # Methods that generate python signal connections
  #-----------------------------------------------------------------------
  # Ports and interfaces will have the same name; their name-mangled
  # counterparts will have a mangled name starting with 'mangled__'.

  def gen_vector_conns( s, d, lhs, rhs, dtype, pos ):
    nbits = dtype.get_length()
    l, r = pos, pos+nbits
    _lhs, _rhs = s._verilator_name(lhs), s._verilator_name(rhs)
    ret = [f"connect( s.{_lhs}, s.mangled__{_rhs}[{l}:{r}] )"]
    return ret, r

  def gen_struct_conns( s, d, lhs, rhs, dtype, pos, symbols ):
    dtype_name = dtype.get_class().__name__
    upblk_name = lhs.replace('.', '_DOT_').replace('[', '_LBR_').replace(']', '_RBR_')
    ret = [
      "@s.update",
      f"def {upblk_name}():",
    ]
    if d == "output":
      ret.append( f"  s.{lhs} = {dtype_name}()" )
    # Patch `dtype_name` into the symbol dictionary
    if dtype_name not in symbols:
      symbols[dtype_name] = dtype.get_class()
    body = []
    all_properties = reversed(list(dtype.get_all_properties().items()))
    for name, field in all_properties:
      _ret, pos = s._gen_dtype_conns( d, lhs+"."+name, rhs, field, pos )
      body += _ret
    return ret + body, pos

  def _gen_vector_conns( s, d, lhs, rhs, dtype, pos ):
    nbits = dtype.get_length()
    l, r = pos, pos+nbits
    _lhs, _rhs = s._verilator_name( lhs ), s._verilator_name( rhs )
    if d == "input":
      ret = [f"  s.mangled__{_rhs}[{l}:{r}] = s.{_lhs}"]
    else:
      ret = [f"  s.{_lhs} = s.mangled__{_rhs}[{l}:{r}]"]
    return ret, r

  def _gen_struct_conns( s, d, lhs, rhs, dtype, pos ):
    ret = []
    all_properties = reversed(list(dtype.get_all_properties().items()))
    for field_name, field in all_properties:
      _ret, pos = s._gen_dtype_conns(d, lhs+"."+field_name, rhs, field, pos)
      ret += _ret
    return ret, pos

  def _gen_packed_array_conns( s, d, lhs, rhs, dtype, n_dim, pos ):
    if not n_dim:
      return s._gen_dtype_conns( d, lhs, rhs, dtype, pos )
    else:
      ret = []
      for idx in range(n_dim[0]):
        _lhs = lhs + f"[{idx}]"
        _ret, pos = \
          s._gen_packed_array_conns( d, _lhs, rhs, dtype, n_dim[1:], pos )
        ret += _ret
      return ret, pos

  def _gen_dtype_conns( s, d, lhs, rhs, dtype, pos ):
    if isinstance( dtype, rdt.Vector ):
      return s._gen_vector_conns( d, lhs, rhs, dtype, pos )
    elif isinstance( dtype, rdt.Struct ):
      return s._gen_struct_conns( d, lhs, rhs, dtype, pos )
    elif isinstance( dtype, rdt.PackedArray ):
      n_dim = dtype.get_dim_sizes()
      sub_dtype = dtype.get_sub_dtype()
      return s._gen_packed_array_conns( d, lhs, rhs, sub_dtype, n_dim, pos )
    else:
      assert False, f"unrecognized data type {dtype}!"

  def gen_dtype_conns( s, d, lhs, rhs, dtype, pos, symbols ):
    if isinstance( dtype, rdt.Vector ):
      return s.gen_vector_conns( d, lhs, rhs, dtype, pos )
    elif isinstance( dtype, rdt.Struct ):
      return s.gen_struct_conns( d, lhs, rhs, dtype, pos, symbols )
    else:
      assert False, f"unrecognized data type {dtype}!"

  def gen_port_conns( s, id_py, id_v, port, n_dim, symbols ):
    if not n_dim:
      d = port.get_direction()
      dtype = port.get_dtype()
      nbits = dtype.get_length()
      ret, pos = s.gen_dtype_conns( d, id_py, id_v, dtype, 0, symbols )
      assert pos == nbits, \
        f"internal error: {id_py} wire length mismatch!"
      return ret
    else:
      ret = []
      for idx in range(n_dim[0]):
        _id_py = id_py + f"[{idx}]"
        _id_v = id_v + f"[{idx}]"
        ret += s.gen_port_conns( _id_py, _id_v, port, n_dim[1:], symbols )
      return ret

  def gen_ifc_conns( s, id_py, id_v, ifc, n_dim, symbols ):
    if not n_dim:
      ret = []
      all_properties = ifc.get_all_properties_packed()
      for name, rtype in all_properties:
        _n_dim, _rtype = s._get_rtype( rtype )
        _id_py = id_py + f".{name}"
        _id_v = id_v + f"__{name}"
        if isinstance( _rtype, rt.Port ):
          ret += s.gen_port_conns( _id_py, _id_v, _rtype, _n_dim, symbols )
        else:
          ret += s.gen_ifc_conns( _id_py, _id_v, _rtype, _n_dim, symbols )
      return ret
    else:
      ret = []
      for idx in range( n_dim[0] ):
        _id_py = id_py + f"[{idx}]"
        _id_v = id_v + f"__{idx}"
        ret += s.gen_ifc_conns( _id_py, _id_v, ifc, n_dim[1:], symbols )
      return ret

  #-------------------------------------------------------------------------
  # gen_signal_decl_py
  #-------------------------------------------------------------------------

  def gen_signal_decl_py( s, rtype ):
    """Return the PyMTL definition of all interface ports of `rtype`."""

    #-----------------------------------------------------------------------
    # Methods that generate signal declarations
    #-----------------------------------------------------------------------

    def gen_dtype_str( symbols, dtype ):
      if isinstance( dtype, rdt.Vector ):
        nbits = dtype.get_length()
        Bits_name = f"Bits{nbits}"
        if Bits_name not in symbols and nbits >= 256:
          Bits_class = mk_bits( nbits )
          symbols.update( { Bits_name : Bits_class } )
        return f'Bits{dtype.get_length()}'
      elif isinstance( dtype, rdt.Struct ):
        # It is possible to reuse the existing struct class because its __init__
        # can be called without arguments.
        name, cls = dtype.get_name(), dtype.get_class()
        if name not in symbols:
          symbols.update( { name : cls } )
        return name
      else:
        assert False, f"unrecognized data type {dtype}!"

    def gen_port_decl_py( ports ):
      symbols, decls = {}, []
      for id_, _port in ports:
        if id_ not in ['clk', 'reset']:
          if isinstance( _port, rt.Array ):
            n_dim = _port.get_dim_sizes()
            rhs = "{direction}( {dtype} )"
            port = _port.get_sub_type()
            _n_dim = copy.copy( n_dim )
            _n_dim.reverse()
            for length in _n_dim:
              rhs = f"[ {rhs} for _ in range({length}) ]"
          else:
            rhs = "{direction}( {dtype} )"
            port = _port
          direction = s._get_direction( port )
          dtype = gen_dtype_str( symbols, port.get_dtype() )
          rhs = rhs.format( **locals() )
          decls.append(f"s.{id_} = {rhs}")
      return symbols, decls

    def gen_ifc_decl_py( ifcs ):

      def gen_ifc_str( symbols, ifc ):

        def _get_arg_str( name, obj ):
          if isinstance( obj, int ):
            return str(obj)
          elif isinstance( obj, Bits ):
            nbits = obj.nbits
            value = int(obj)
            Bits_name = f"Bits{nbits}"
            Bits_arg_str = f"{Bits_name}( {value} )"
            if Bits_name not in symbols and nbits >= 256:
              Bits_class = mk_bits( nbits )
              symbols.update( { Bits_name : Bits_class } )
            return Bits_arg_str
          elif is_bitstruct_inst( obj ):
            raise TypeError("Do you really want to pass in an instance of "
                            "a BitStruct? Contact PyMTL developers!")
            # This is hacky: we don't know how to construct an object that
            # is the same as `obj`, but we do have the object itself. If we
            # add `obj` to the namespace of `construct` everything works fine
            # but the user cannot tell what object is passed to the constructor
            # just from the code.
            # Do not use a double underscore prefix because that will be
            # interpreted differently by the Python interpreter
            # bs_name = ("_" if name[0] != "_" else "") + name + "_obj"
            # if bs_name not in symbols:
              # symbols.update( { bs_name : obj } )
            # return bs_name
          elif isinstance( obj, type ) and issubclass( obj, Bits ):
            nbits = obj.nbits
            Bits_name = f"Bits{nbits}"
            if Bits_name not in symbols and nbits >= 256:
              Bits_class = mk_bits( nbits )
              symbols.update( { Bits_name : Bits_class } )
            return Bits_name
          elif is_bitstruct_class(obj):
            BitStruct_name = obj.__name__
            if BitStruct_name not in symbols:
              symbols.update( { BitStruct_name : obj } )
            return BitStruct_name

          raise TypeError( f"Interface constructor argument {obj} is not an int/Bits/BitStruct!" )

        name, cls = ifc.get_name(), ifc.get_class()
        if name not in symbols:
          symbols.update( { name : cls } )
        arg_list = []
        args = ifc.get_args()
        for idx, obj in enumerate(args[0]):
          arg_list.append( _get_arg_str( f"_ifc_arg{idx}", obj ) )
        for arg_name, arg_obj in args[1].items():
          arg_list.append( f"{arg_name} = {_get_arg_str( arg_name, arg_obj )}" )
        return name, ', '.join( arg_list )

      symbols, decls = {}, []
      for id_, ifc in ifcs:
        if isinstance( ifc, rt.Array ):
          n_dim = ifc.get_dim_sizes()
          rhs = "{ifc_class}({ifc_params})"
          _ifc = ifc.get_sub_type()
          _n_dim = copy.copy( n_dim )
          _n_dim.reverse()
          for length in _n_dim:
            rhs = f"[ {rhs} for _ in range({length}) ]"
        else:
          rhs = "{ifc_class}({ifc_params})"
          _ifc = ifc
        ifc_class, ifc_params = gen_ifc_str( symbols, _ifc )
        if ifc_params:
          ifc_params = " " + ifc_params + " "
        rhs = rhs.format( **locals() )
        decls.append(f"s.{id_} = {rhs}")
      return symbols, decls

    #-----------------------------------------------------------------------
    # Method gen_signal_decl_py
    #-----------------------------------------------------------------------

    ports = rtype.get_ports_packed()
    ifcs = rtype.get_ifc_views_packed()

    p_symbols, p_decls = gen_port_decl_py( ports )
    i_symbols, i_decls = gen_ifc_decl_py( ifcs )

    p_conns, i_conns = [], []
    struct_conn_symbols = {}
    for id_, port in ports:
      p_n_dim, p_rtype = s._get_rtype( port )
      p_conns += s.gen_port_conns( id_, id_, p_rtype, p_n_dim, struct_conn_symbols )
    for id_, ifc in ifcs:
      i_n_dim, i_rtype = s._get_rtype( ifc )
      i_conns += s.gen_ifc_conns( id_, id_, i_rtype, i_n_dim, struct_conn_symbols )

    p_symbols.update( i_symbols )
    p_symbols.update( struct_conn_symbols )
    decls = p_decls + i_decls
    conns = p_conns + i_conns

    return p_symbols, decls, conns

  #-------------------------------------------------------------------------
  # gen_wire_decl_py
  #-------------------------------------------------------------------------

  def gen_wire_decl_py( s, name, _wire ):
    """Return the PyMTL definition of `wire`."""
    template = "s.mangled__{name} = {rhs}"
    rhs = "Wire( Bits{nbits} )"
    name = s._verilator_name( name )
    n_dim, rtype = s._get_rtype( _wire )
    dtype = rtype.get_dtype()
    nbits = dtype.get_length()
    for idx in reversed(n_dim):
      rhs = f"[ {rhs} for _ in range({idx}) ]"
    rhs = rhs.format( **locals() )
    return template.format( **locals() )

  #-------------------------------------------------------------------------
  # gen_comb_input
  #-------------------------------------------------------------------------

  def gen_port_array_input( s, lhs, rhs, dtype, n_dim ):
    nbits = dtype.get_length()
    if not n_dim:
      return s._gen_ref_write( lhs, rhs, nbits )
    else:
      ret = []
      for idx in range( n_dim[0] ):
        _lhs = f"{lhs}[{idx}]"
        _rhs = f"{rhs}[{idx}]"
        ret += s.gen_port_array_input( _lhs, _rhs, dtype, n_dim[1:] )
      return ret

  def gen_comb_input( s, packed_ports ):
    ret = []
    # Read all input ports ( except for 'clk' ) from component ports into
    # the verilated model. We do NOT want `clk` signal to be read into
    # the verilated model because only the sequential update block of
    # the imported component should manipulate it.
    for py_name, v_name, rtype in packed_ports:
      p_n_dim, p_rtype = s._get_rtype( rtype )
      if s._get_direction( p_rtype ) == 'InPort' and py_name != 'clk' and v_name:
        dtype = p_rtype.get_dtype()
        lhs = "_ffi_m."+s._verilator_name(v_name)
        rhs = "s.mangled__"+s._verilator_name(py_name)
        ret += s.gen_port_array_input( lhs, rhs, dtype, p_n_dim )
    return ret

  #-------------------------------------------------------------------------
  # gen_comb_output
  #-------------------------------------------------------------------------

  def gen_port_array_output( s, lhs, rhs, dtype, n_dim ):
    nbits = dtype.get_length()
    if not n_dim:
      return s._gen_ref_read( lhs, rhs, nbits )
    else:
      ret = []
      for idx in range( n_dim[0] ):
        _lhs = f"{lhs}[{idx}]"
        _rhs = f"{rhs}[{idx}]"
        ret += s.gen_port_array_output( _lhs, _rhs, dtype, n_dim[1:] )
      return ret

  def gen_comb_output( s, packed_ports ):
    ret = []
    for py_name, v_name, rtype in packed_ports:
      p_n_dim, p_rtype = s._get_rtype( rtype )
      if s._get_direction( rtype ) == 'OutPort':
        dtype = p_rtype.get_dtype()
        lhs = "s.mangled__" + s._verilator_name(py_name)
        rhs = "_ffi_m." + s._verilator_name(v_name)
        ret += s.gen_port_array_output( lhs, rhs, dtype, p_n_dim )
    return ret

  #-------------------------------------------------------------------------
  # gen_constraints
  #-------------------------------------------------------------------------

  def _gen_constraints( s, name, n_dim, rtype ):
    if not n_dim:
      return [f"U( seq_upblk ) < RD( s.mangled__{name} ),"]
    else:
      ret = []
      for i in range( n_dim[0] ):
        ret += s._gen_constraints( f"{name}[{i}]", n_dim[1:], rtype )
      return ret

  def gen_constraints( s, packed_ports ):
    ret = []
    for py_name, v_name, rtype in packed_ports:
      if s._get_direction( rtype ) == 'OutPort':
        if isinstance( rtype, rt.Array ):
          n_dim = rtype.get_dim_sizes()
          sub_type = rtype.get_sub_type()
          ret += s._gen_constraints( py_name, n_dim, sub_type )
        else:
          wire_name = "s.mangled__" + s._verilator_name(py_name)
          ret.append( f"U( seq_upblk ) < RD( {wire_name} )," )
    ret.append( "U( seq_upblk ) < U( comb_upblk )," )
    return ret

  #-------------------------------------------------------------------------
  # gen_line_trace_py
  #-------------------------------------------------------------------------

  def gen_line_trace_py( s, packed_ports ):
    """Return the line trace method body that shows all interface ports."""
    ret = [ 'lt = ""' ]
    template = 'lt += "{my_name} = {{}}, ".format({full_name})'
    for name, v_name, port in packed_ports:
      my_name = name
      full_name = 's.mangled__'+s._verilator_name(name)
      ret.append( template.format( **locals() ) )
    ret.append( 'return lt' )
    make_indent( ret, 3 )
    return '\n'.join( ret )

  #-------------------------------------------------------------------------
  # gen_internal_line_trace_py
  #-------------------------------------------------------------------------

  def gen_internal_line_trace_py( s, packed_ports ):
    """Return the line trace method body that shows all CFFI ports."""
    ret = [ '_ffi_m = s._ffi_m', 'lt = ""' ]
    template = \
      "lt += '{v_name} = {{}}, '.format(full_vector(s.mangled__{my_name}, _ffi_m.{v_name}))"
    for my_name, v_name, port in packed_ports:
      if v_name:
        my_name = s._verilator_name(my_name)
        v_name = s._verilator_name(v_name)
        ret.append( template.format(**locals()) )
    ret.append( 'return lt' )
    make_indent( ret, 2 )
    return '\n'.join( ret )

  #=========================================================================
  # Helper functions
  #=========================================================================

  def _verilator_name( s, name ):
    # TODO: PyMTL translation should generate dollar-sign-free Verilog source
    # code. Verify that this replacement rule here is not necessary.
    return name.replace('__', '___05F').replace('$', '__024')

  def _get_direction( s, port ):
    if isinstance( port, rt.Port ):
      d = port.get_direction()
    elif isinstance( port, rt.Array ):
      d = port.get_sub_type().get_direction()
    else:
      assert False, f"{port} is not a port or array of ports!"
    if d == 'input':
      return 'InPort'
    elif d == 'output':
      return 'OutPort'
    else:
      assert False, f"unrecognized direction {d}!"

  def _get_c_n_dim( s, port ):
    if isinstance( port, rt.Array ):
      return port.get_dim_sizes()
    else:
      return []

  def _get_c_dim( s, port ):
    return "".join( f"[{i}]" for i in s._get_c_n_dim(port) )

  def _get_c_nbits( s, port ):
    if isinstance( port, rt.Array ):
      dtype = port.get_sub_type().get_dtype()
    else:
      dtype = port.get_dtype()
    return dtype.get_length()

  def _gen_ref_write( s, lhs, rhs, nbits ):
    if nbits <= 64:
      return [ f"{lhs}[0] = int({rhs})" ]
    else:
      ret = []
      ITEM_BITWIDTH = 32
      num_assigns = (nbits-1)//ITEM_BITWIDTH+1
      for idx in range(num_assigns):
        l = ITEM_BITWIDTH*idx
        r = l+ITEM_BITWIDTH if l+ITEM_BITWIDTH <= nbits else nbits
        ret.append( f"{lhs}[{idx}] = int({rhs}[{l}:{r}])" )
      return ret

  def _gen_ref_read( s, lhs, rhs, nbits ):
    if nbits <= 64:
      return [ f"{lhs} = Bits{nbits}({rhs}[0])" ]
    else:
      ret = []
      ITEM_BITWIDTH = 32
      num_assigns = (nbits-1)//ITEM_BITWIDTH+1
      for idx in range(num_assigns):
        l = ITEM_BITWIDTH*idx
        r = l+ITEM_BITWIDTH if l+ITEM_BITWIDTH <= nbits else nbits
        _nbits = r - l
        ret.append( f"{lhs}[{l}:{r}] = Bits{_nbits}({rhs}[{idx}])" )
      return ret

  def _get_rtype( s, _rtype ):
    if isinstance( _rtype, rt.Array ):
      n_dim = _rtype.get_dim_sizes()
      rtype = _rtype.get_sub_type()
    else:
      n_dim = []
      rtype = _rtype
    return n_dim, rtype



#=========================================================================
# SVStructuralTranslatorL4.py
#=========================================================================
"""Provide SystemVerilog structural translator implementation."""

from pymtl3.passes.backends.generic.structural.StructuralTranslatorL4 import (
    StructuralTranslatorL4,
)
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt

from ...util.utility import make_indent
from .SVStructuralTranslatorL3 import SVStructuralTranslatorL3


class SVStructuralTranslatorL4(
    SVStructuralTranslatorL3, StructuralTranslatorL4 ):

  #-----------------------------------------------------------------------
  # Declarations
  #-----------------------------------------------------------------------

  def rtlir_tr_subcomp_port_decls( s, _port_decls ):
    port_defs = [x['def'] for x in _port_decls]
    port_decls = [x['decl'] for x in _port_decls]
    make_indent( port_defs, 1 )
    make_indent( port_decls, 2 )
    return {
      'def' : '\n'.join( port_defs ),
      'decl' : ',\n'.join( port_decls )
    }

  def rtlir_tr_subcomp_port_decl( s, m, c_id, c_rtype, c_array_type, port_id,
      port_rtype, port_dtype, port_array_type ):
    port_def_rtype = rt.Wire(port_dtype["raw_dtype"])

    return {
      'def' : s.rtlir_tr_wire_decl('{c_id}__'+port_id, port_def_rtype,
                port_array_type, port_dtype),
      'decl' : f'.{port_id}( {{c_id}}__{port_id} )'
    }

  def rtlir_tr_subcomp_ifc_port_decls( s, _ifc_port_decls ):
    port_defs = [x['def'] for x in _ifc_port_decls]
    port_decls = [x['decl'] for x in _ifc_port_decls]
    make_indent( port_defs, 1 )
    make_indent( port_decls, 2 )
    return {
      'def' : '\n'.join( port_defs ),
      'decl' : ',\n'.join( port_decls )
    }

  def rtlir_tr_subcomp_ifc_port_decl( s, m, c_id, c_rtype, c_array_type,
      ifc_id, ifc_rtype, ifc_array_type, port_id, port_rtype,
      port_array_type ):
    assert isinstance( port_rtype, rt.Port ), \
      f"SystemVerilog backend does not support nested interface {port_id} yet!"
    port_dtype = s.rtlir_data_type_translation( m, port_rtype.get_dtype() )
    port_def_rtype = rt.Wire(port_dtype["raw_dtype"])

    return {
      'def' : s.rtlir_tr_wire_decl('{c_id}__{ifc_id}__'+port_id, port_def_rtype,
                port_array_type, port_dtype),
      'decl' : f'.{{ifc_id}}__{port_id}( {{c_id}}__{{ifc_id}}__{port_id} )'
    }

  def rtlir_tr_subcomp_ifc_decls( s, _ifc_decls ):
    _ifc_decls = sum( _ifc_decls, [] )
    ifc_defs = [x['def'] for x in _ifc_decls]
    ifc_decls = [x['decl'] for x in _ifc_decls]
    return {
      'def' : '\n'.join( ifc_defs ),
      'decl' : ',\n'.join( ifc_decls )
    }

  def rtlir_tr_subcomp_ifc_decl( s, m, c_id, c_rtype, c_array_type,
      ifc_id, ifc_rtype, ifc_array_type, ports ):

    def gen_subcomp_ifc_decl( ifc_id, ifc_rtype, n_dim, c_n_dim, ports ):
      if not n_dim:
        c_id = '{c_id}'
        ifc_id = ifc_id + c_n_dim
        return [ {
          'def' : ports['def'].format( **locals() ),
          'decl' : ports['decl'].format( **locals() )
        } ]
      else:
        return sum( [gen_subcomp_ifc_decl( ifc_id, ifc_rtype, n_dim[1:],
            c_n_dim+'__'+str( idx ), ports ) for idx in range( n_dim[0] )], [] )

    n_dim = ifc_array_type[ 'n_dim' ]
    return \
      gen_subcomp_ifc_decl( ifc_id, ifc_rtype, n_dim, "", ports )

  def rtlir_tr_subcomp_decls( s, subcomps ):
    subcomp_decls = sum( subcomps, [] )
    return '\n\n'.join( subcomp_decls )

  def rtlir_tr_subcomp_decl( s, m, c_id, c_rtype, c_array_type, port_conns, ifc_conns ):

    _c_name = s.rtlir_tr_component_unique_name( c_rtype )

    def gen_subcomp_array_decl( c_id, port_conns, ifc_conns, n_dim, c_n_dim ):
      tplt = \
"""\
{port_wire_defs}{ifc_inst_defs}

  {c_name} {c_id}
  (
{port_conn_decls}{ifc_conn_decls}
  );\
"""
      if not n_dim:
        # Add the component dimension to the defs/decls
        c_id = c_id + c_n_dim
        c_name = _c_name
        port_wire_defs = port_conns['def'].format( **locals() )
        ifc_inst_defs = ifc_conns['def'].format( **locals() )
        if port_wire_defs and ifc_inst_defs:
          port_wire_defs += '\n'
        port_conn_decls = port_conns['decl'].format( **locals() )
        ifc_conn_decls = ifc_conns['decl'].format( **locals() )
        if port_conn_decls and ifc_conn_decls:
          port_conn_decls += '\n'
        return [ tplt.format( **locals() ) ]

      else:
        return sum( [gen_subcomp_array_decl( c_id,
            port_conns, ifc_conns, n_dim[1:], c_n_dim+'__'+str(idx) ) \
                for idx in range( n_dim[0] )], [] )

    # If `c_array_type` is not None we need to impelement an array of
    # components, each with their own connections for the ports.
    # This means we will only support component indexing where the index
    # is a constant integer.
    n_dim = c_array_type['n_dim']
    if port_conns['decl'] and ifc_conns['decl']:
      port_conns['decl'] += ','
    return\
      gen_subcomp_array_decl( c_id, port_conns, ifc_conns, n_dim, '' )

  #-----------------------------------------------------------------------
  # Signal operations
  #-----------------------------------------------------------------------

  def rtlir_tr_component_array_index( s, base_signal, index ):
    return f'{base_signal}__{index}'

  def rtlir_tr_subcomp_attr( s, base_signal, attr ):
    return f'{base_signal}__{attr}'



"""
==========================================================================
 xcel_ifcs_test.py
==========================================================================

Author : Yanghui Ou
  Date : June 4, 2019
"""

from pymtl3 import *
from pymtl3.stdlib.ifcs import XcelMsgType, mk_xcel_msg
from pymtl3.stdlib.rtl import RegisterFile
from pymtl3.stdlib.rtl.queues import NormalQueueRTL

from .xcel_ifcs import (
    XcelMasterIfcCL,
    XcelMasterIfcFL,
    XcelMasterIfcRTL,
    XcelMinionIfcCL,
    XcelMinionIfcFL,
    XcelMinionIfcRTL,
)

#-------------------------------------------------------------------------
# FL master/minion
#-------------------------------------------------------------------------

class SomeMasterNonBlockingFL( Component ):

  # Actually we don't need ReqType and RespType for FL models. It's just
  # for polymorphic instantiation in the test harness, since CL/RTL models
  # have these two parameters.
  def construct( s, ReqType, RespType, nregs=16 ):
    s.xcel = XcelMasterIfcFL()

    s.addr = 0
    s.nregs = nregs
    s.trace = ""

    @s.update
    def up_master_while():
      while s.addr < s.nregs:
        s.trace = "#            "
        wr_data = 0xbabe0000 | s.addr
        s.xcel.write( s.addr, wr_data )
        s.trace = "wr:{:x}:{:x}".format( int(s.addr), int(wr_data) )
        rd_data = s.xcel.read( s.addr )
        assert rd_data == wr_data, "{} {}".format( hex(int(rd_data)), hex(int(wr_data)) )
        s.trace = "rd:{:x}:{:x}".format( int(s.addr), int(rd_data) )
        s.addr += 1

  def done( s ):
    return s.addr >= s.nregs

  def line_trace( s ):
    ret = s.trace
    s.trace = "#            "
    return ret

class SomeMasterBlockingFL( Component ):

  def construct( s, ReqType, RespType, nregs=16 ):
    s.xcel = XcelMasterIfcFL()

    s.addr = 0
    s.nregs = nregs
    s.trace = ""

    @s.update
    def up_master_noloop():
      if s.addr < s.nregs:
        s.trace = "#            "
        wr_data = 0xbabe0000 | s.addr
        s.xcel.write( s.addr, wr_data )
        s.trace = "wr:{:x}:{:x}".format( s.addr, wr_data )
        rd_data = s.xcel.read( s.addr )
        assert rd_data == wr_data, "{} {}".format( hex(int(rd_data)), hex(int(wr_data)) )
        s.trace = "rd:{:x}:{:x}".format( int(s.addr), int(rd_data) )
        s.addr += 1

  def done( s ):
    return s.addr >= s.nregs

  def line_trace( s ):
    ret = s.trace
    s.trace = "#            "
    return ret

class SomeMinionFL( Component ):

  def read_method( s, addr ):
    return s.reg_file[ int(addr) ]

  def write_method( s, addr, data ):
    s.reg_file[ int(addr) ] = data

  def construct( s, ReqType, RespType, nregs=16 ):
    s.xcel = XcelMinionIfcFL( read=s.read_method, write=s.write_method )
    s.reg_file = [ 0 for _ in range( nregs ) ]

  def line_trace( s ):
    return s.xcel.line_trace()

#-------------------------------------------------------------------------
# CL master/minion
#-------------------------------------------------------------------------

class SomeMasterCL( Component ):

  def recv( s, msg ):
    if msg.type_ == XcelMsgType.READ:
      assert msg.data == 0xface0000 | s.addr-1
      s.count += 1

  def recv_rdy( s ):
    return True

  def construct( s, ReqType, RespType, nregs=16 ):
    s.xcel = XcelMasterIfcCL( ReqType, RespType, resp=s.recv, resp_rdy=s.recv_rdy )
    s.addr = 0
    s.count = 0
    s.nregs = nregs
    s.flag = True

    DataType = ReqType.get_field_type( 'data' )
    assert DataType is RespType.get_field_type( 'data' )
    AddrType = ReqType.get_field_type( 'addr' )

    @s.update
    def up_master_req():
      if s.xcel.req.rdy():
        if s.flag:
          s.xcel.req( ReqType( XcelMsgType.WRITE, AddrType(s.addr),
                               DataType(0xface0000 | s.addr) ) )
          s.flag = not s.flag
        else:
          s.xcel.req( ReqType( XcelMsgType.READ, AddrType(s.addr),
                               DataType(0) ) )
          s.addr += 1
          s.flag = not s.flag

  def done( s ):
    return s.count == s.nregs

  def line_trace( s ):
    return "{}".format( s.xcel )

class SomeMinionCL( Component ):

  def recv( s, msg ):
    assert s.entry is None
    s.entry = msg

  def recv_rdy( s ):
    return s.entry is None

  def read( s, addr ):
    addr = int(addr)
    return s.reg_file[ addr ]

  def write( s, addr, data ):
    addr = int(addr)
    s.reg_file[ addr ] = data

  def construct( s, ReqType, RespType, nregs=16 ):
    s.xcel = XcelMinionIfcCL( ReqType, RespType, req=s.recv, req_rdy=s.recv_rdy )
    s.entry = None
    s.reg_file = [ 0 for _ in range( nregs ) ]

    @s.update
    def up_process():

      if s.entry is not None and s.xcel.resp.rdy():
        # Dequeue xcel request message
        req     = s.entry
        s.entry = None

        if req.type_ == XcelMsgType.READ:
          resp = RespType( req.type_, s.read( req.addr ) )

        elif req.type_ == XcelMsgType.WRITE:
          s.write( req.addr, req.data )
          resp = RespType( req.type_, 0 )

        s.xcel.resp( resp )

    s.add_constraints( U(up_process) < M(s.xcel.req) ) # pipeline behavior

  def line_trace( s ):
    return "{}".format( s.xcel )

#-------------------------------------------------------------------------
# RTL master/minion
#-------------------------------------------------------------------------

class SomeMasterRTL( Component ):

  def construct( s, ReqType, RespType, nregs=16 ):

    # Interface

    s.xcel = XcelMasterIfcRTL( ReqType, RespType )

    # Local parameters

    DataType = ReqType.get_field_type( 'data' )
    assert DataType is RespType.get_field_type( 'data' )
    AddrType = ReqType.get_field_type( 'addr' )

    s.nregs = nregs

    # Components

    s.addr  = Wire( AddrType )
    s.count = Wire( Bits16   )
    s.flag  = Wire( Bits1    )

    @s.update_ff
    def up_rtl_addr():
      if s.reset:
        s.addr <<= AddrType(0)
      elif s.xcel.req.en and not s.flag:
        s.addr <<= s.addr + AddrType(1)

    @s.update_ff
    def up_rtl_flag():
      if s.reset:
        s.flag <<= Bits1(1)
      elif s.xcel.req.en:
        s.flag <<= ~s.flag

    @s.update_ff
    def up_rtl_count():
      if s.reset:
        s.count <<= Bits16(0)
      elif s.xcel.resp.en and s.xcel.resp.msg.type_ == XcelMsgType.READ:
        s.count <<= s.count + Bits16(1)

    @s.update
    def up_req():
      s.xcel.req.en = s.xcel.req.rdy if ~s.reset else Bits1(0)
      s.xcel.req.msg.type_ = XcelMsgType.WRITE if s.flag else XcelMsgType.READ
      s.xcel.req.msg.addr = s.addr
      s.xcel.req.msg.data = DataType( 0xface0000 | int(s.addr) )

    @s.update
    def up_resp():
      s.xcel.resp.rdy = Bits1(1)

  def done( s ):
    return s.count == s.nregs

  def line_trace( s ):
    return "{}({}){}".format( s.xcel.req, s.flag, s.xcel.resp )

class SomeMinionRTL( Component ):

  def construct( s, ReqType, RespType, nregs=16 ):

    # Interface

    s.xcel = XcelMinionIfcRTL( ReqType, RespType )

    # Local parameters

    DataType = ReqType.get_field_type( 'data' )
    assert DataType is RespType.get_field_type( 'data' )

    s.nregs = nregs

    # Components

    s.req_q = NormalQueueRTL( ReqType, num_entries=1 )
    s.wen   = Wire( Bits1 )
    s.reg_file = RegisterFile( DataType, nregs )(
      raddr = { 0: s.req_q.deq.ret.addr },
      rdata = { 0: s.xcel.resp.msg.data },
      wen   = { 0: s.wen                },
      waddr = { 0: s.req_q.deq.ret.addr },
      wdata = { 0: s.req_q.deq.ret.data },
    )
    connect( s.xcel.req,            s.req_q.enq           )
    connect( s.xcel.resp.msg.type_, s.req_q.deq.ret.type_ )

    @s.update
    def up_wen():
      s.wen = s.req_q.deq.rdy and s.req_q.deq.ret.type_ == XcelMsgType.WRITE

    @s.update
    def up_resp():
      s.xcel.resp.en = s.req_q.deq.rdy and s.xcel.resp.rdy
      s.req_q.deq.en = s.req_q.deq.rdy and s.xcel.resp.rdy

  def line_trace( s ):
    return "{}".format( s.xcel )

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s,
                 MasterType = SomeMasterBlockingFL,
                 MinionType=SomeMinionFL,
                 nregs = 16 ):
    ReqType, RespType = mk_xcel_msg( addr=clog2(nregs), data=32 )
    s.master = MasterType( ReqType, RespType, nregs=nregs )
    s.minion = MinionType( ReqType, RespType, nregs=nregs )

    connect( s.master.xcel, s.minion.xcel )

  def line_trace( s ):
    return "{} > {}".format( s.master.line_trace(), s.minion.line_trace() )

  def done( s ):
    return s.master.done()

  def run_sim( s, max_cycles=1000 ):
    s.apply( SimulationPass() )
    # Run simulation
    print("")
    ncycles = 0
    s.sim_reset()
    print("{:3}: {}".format( ncycles, s.line_trace() ))
    while not s.done() and ncycles < max_cycles:
      s.tick()
      ncycles += 1
      print("{:3}: {}".format( ncycles, s.line_trace() ))

    # Check timeout
    assert ncycles < max_cycles

#-------------------------------------------------------------------------
# FL-FL composition
#-------------------------------------------------------------------------

def test_xcel_fl_fl_blocking():
  th = TestHarness()
  th.set_param( "top.construct",
    MasterType = SomeMasterBlockingFL,
    MinionType = SomeMinionFL,
    nregs      = 16,
  )
  th.run_sim()

def test_xcel_fl_fl_nonblocking():
  th = TestHarness()
  th.set_param( "top.construct",
    MasterType = SomeMasterNonBlockingFL,
    MinionType = SomeMinionFL,
    nregs      = 16,
  )
  th.run_sim()

#-------------------------------------------------------------------------
# FL-CL composition
#-------------------------------------------------------------------------

def test_xcel_fl_cl_blocking():
  th = TestHarness()
  th.set_param( "top.construct",
    MasterType = SomeMasterBlockingFL,
    MinionType = SomeMinionCL,
    nregs      = 16,
  )
  th.run_sim()

def test_xcel_fl_cl_nonblocking():
  th = TestHarness()
  th.set_param( "top.construct",
    MasterType = SomeMasterNonBlockingFL,
    MinionType = SomeMinionCL,
    nregs      = 16,
  )
  th.run_sim()

#-------------------------------------------------------------------------
# FL-RTL composition
#-------------------------------------------------------------------------

def test_xcel_fl_rtl_blocking():
  th = TestHarness()
  th.set_param( "top.construct",
    MasterType = SomeMasterBlockingFL,
    MinionType = SomeMinionRTL,
    nregs      = 16,
  )
  th.run_sim()

def test_xcel_fl_rtl_nonblocking():
  th = TestHarness()
  th.set_param( "top.construct",
    MasterType = SomeMasterNonBlockingFL,
    MinionType = SomeMinionRTL,
    nregs      = 16,
  )
  th.run_sim()

#-------------------------------------------------------------------------
# CL-CL composition
#-------------------------------------------------------------------------

def test_xcel_cl_cl():
  th = TestHarness()
  th.set_param( "top.construct",
    MasterType = SomeMasterCL,
    MinionType = SomeMinionCL,
    nregs      = 8,
  )
  th.run_sim()

#-------------------------------------------------------------------------
# CL-RTL composition
#-------------------------------------------------------------------------

def test_xcel_cl_rtl():
  th = TestHarness()
  th.set_param( "top.construct",
    MasterType = SomeMasterCL,
    MinionType = SomeMinionRTL,
    nregs      = 8,
  )
  th.run_sim()

#-------------------------------------------------------------------------
# CL-FL composition
#-------------------------------------------------------------------------

def test_xcel_cl_fl():
  th = TestHarness()
  th.set_param( "top.construct",
    MasterType = SomeMasterCL,
    MinionType = SomeMinionFL,
    nregs      = 8,
  )
  th.run_sim()

#-------------------------------------------------------------------------
# RTL-RTL composition
#-------------------------------------------------------------------------

def test_xcel_rtl_rtl():
  th = TestHarness()
  th.set_param( "top.construct",
    MasterType = SomeMasterRTL,
    MinionType = SomeMinionRTL,
    nregs      = 8,
  )
  th.run_sim()

#-------------------------------------------------------------------------
# RTL-CL composition
#-------------------------------------------------------------------------

def test_xcel_rtl_cl():
  th = TestHarness()
  th.set_param( "top.construct",
    MasterType = SomeMasterRTL,
    MinionType = SomeMinionCL,
    nregs      = 8,
  )
  th.run_sim()

#-------------------------------------------------------------------------
# RTL-FL composition
#-------------------------------------------------------------------------

def test_xcel_rtl_fl():
  th = TestHarness()
  th.set_param( "top.construct",
    MasterType = SomeMasterRTL,
    MinionType = SomeMinionFL,
    nregs      = 8,
  )
  th.run_sim()



"""
========================================================================
CollectSignalPass.py
========================================================================
collects signals and stored it as _collect_signals attribute of top.

Author : Kaishuo Cheng, Shunning Jiang
Date   : Nov 9, 2019
"""

from collections import defaultdict

import py

from pymtl3.datatypes import Bits, is_bitstruct_class, to_bits
from pymtl3.dsl import Const
from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.errors import ModelTypeError, PassOrderError


class CollectSignalPass( BasePass ):
  def __call__( self, top ):
    if hasattr( top, "config_tracing" ):
      top.config_tracing.check()

      if top.config_tracing.tracing in [ 'text_ascii', 'text_fancy' ]:
        if not hasattr( top, "_tracing" ):
          top._tracing = PassMetadata()
        top._tracing.collect_text_sigs = self.collect_sig_func( top, top._tracing )

  def collect_sig_func( self, top, wavmeta ):

    # TODO use actual nets to reduce the amount of saved signals

    # Give all ' and " characters a preceding backslash for .format
    wav_srcs = []

    # Now we create per-cycle signal value collect functions
    for x in top._dsl.all_signals:
      if x.is_top_level_signal() and ( not repr(x).endswith('.clk') or x is top.clk ):
        if is_bitstruct_class( x._dsl.Type ):
          wav_srcs.append( "wavmeta.text_sigs['{0}'].append( to_bits({0}).bin() )".format(x) )
        elif issubclass( x._dsl.Type, Bits ):
          wav_srcs.append( "wavmeta.text_sigs['{0}'].append( {0}.bin() )".format(x) )

    wavmeta.text_sigs = defaultdict(list)

    # TODO use integer index instead of dict, should be easy
    src =  """
def dump_wav():
  {}
""".format( "\n  ".join(wav_srcs) )
    s, l_dict = top, {}
    exec(compile( src, filename="temp", mode="exec"), globals().update(locals()), l_dict)
    return l_dict['dump_wav']



"""
========================================================================
GenDAGPass.py
========================================================================
Generate a DAG of update blocks (including net connection blocks) from
a model.

Author : Shunning Jiang
Date   : Jan 18, 2018
"""
from collections import defaultdict, deque
from linecache import cache as line_cache

from pymtl3.datatypes import *
from pymtl3.datatypes.helpers import get_bitstruct_inst_all_classes
from pymtl3.dsl import *
from pymtl3.dsl.errors import LeftoverPlaceholderError
from pymtl3.passes.BasePass import BasePass, PassMetadata


class GenDAGPass( BasePass ):

  def __call__( self, top ):
    top.check()
    top._dag = PassMetadata()

    placeholders = [ x for x in top._dsl.all_named_objects
                     if isinstance( x, Placeholder ) ]

    if placeholders:
      raise LeftoverPlaceholderError( placeholders )

    self._generate_net_blocks( top )
    self._process_value_constraints( top )
    self._process_methods( top )

  def _generate_net_blocks( self, top ):
    """ _generate_net_blocks:
    Each net is an update block. Readers are actually "written" here.
      >>> s.net_reader1 = s.net_writer
      >>> s.net_reader2 = s.net_writer """

    top._dag.genblks = set()
    top._dag.genblk_hostobj = {}
    top._dag.genblk_reads   = {}
    top._dag.genblk_writes  = {}
    # top._dag.genblk_src     = {}

    # Fall back to compiling one block at a time
    # This is currently because there might be different structs with
    # the same name but essentially different type. It requires name
    # disambiguation to let them co-exist in closure. With block-by-block
    # compilation, we minimize the effect.

    # TODO see if directly compiling AST instead of source can be faster
    def compile_net_blk( _globals, src ):
      _locals = {}
      fname = f"Net at {_globals['s']!r}"
      exec( compile( src, filename=fname, mode="exec"), _globals, _locals )
      line_cache[ fname ] = (len(src), None, src.splitlines(), fname )
      return list(_locals.values())[0]

    for writer, signals in top.get_all_value_nets():
      if len(signals) == 1:
        continue

      readers = [ x for x in signals if x is not writer ]
      fanout  = len( readers )

      wr_lca  = writer.get_host_component()
      rd_lcas = [ x.get_host_component() for x in readers ]

      # Find common ancestor: iteratively go to parent level and check if
      # at the same level all objects' ancestors are the same

      mindep  = min( wr_lca.get_component_level(),
                     min( [ x.get_component_level() for x in rd_lcas ] ) )

      # First navigate all objects to the same level deep

      for i in range( mindep, wr_lca.get_component_level() ):
        wr_lca = wr_lca.get_parent_object()

      for i, x in enumerate( rd_lcas ):
        for j in range( mindep, x.get_component_level() ):
          x = x.get_parent_object()
        rd_lcas[i] = x

      # Then iteratively check if their ancestor is the same

      while wr_lca is not top:
        succeed = True
        for x in rd_lcas:
          if x is not wr_lca:
            succeed = False
            break
        if succeed: break

        # Bring up all objects for another level
        wr_lca = wr_lca.get_parent_object()
        for i in range( fanout ):
          rd_lcas[i] = rd_lcas[i].get_parent_object()

      lca_len = len( repr(wr_lca) )
      _globals = {'s': wr_lca }

      if isinstance( writer, Const ) and type(writer._dsl.const) is not int:
        types = get_bitstruct_inst_all_classes( writer._dsl.const )

        for t in types:
          if t.__name__ in _globals:
            assert t is _globals[ t.__name__ ], "Cannot handle two subfields with the same struct name but different structs"
          _globals[ t.__name__ ] = t
        wstr = repr(writer)

      else:
        wstr = f"s.{repr(writer)[lca_len+1:]}"

      rstrs   = [ f"s.{repr(x)[lca_len+1:]}" for x in readers ]
      upblk_name = f"{writer!r}__{fanout}".replace( " ", "" ) \
                      .replace( ".", "_" ).replace( ":", "_" ) \
                      .replace( "[", "_" ).replace( "]", "_" ) \
                      .replace( "(", "_" ).replace( ")", "_" ) \
                      .replace( ",", "_" )

      gen_src = """
def {}():
  {} = {}""".format( upblk_name, " = ".join( rstrs ), wstr )

      blk = compile_net_blk( _globals, gen_src )

      top._dag.genblks.add( blk )
      if writer.is_signal():
        top._dag.genblk_reads[ blk ] = [ writer ]
      top._dag.genblk_writes[ blk ] = readers

    # Get the final list of update blocks
    top._dag.final_upblks = top.get_all_update_blocks() | top._dag.genblks

  def _process_value_constraints( self, top ):

    # Query update block metadata from top

    update_ff                    = top.get_all_update_ff()
    upblk_reads, upblk_writes, _ = top.get_all_upblk_metadata()
    genblk_reads, genblk_writes  = top._dag.genblk_reads, top._dag.genblk_writes
    U_U, RD_U, WR_U, U_M         = top.get_all_explicit_constraints()

    #---------------------------------------------------------------------
    # Explicit constraint
    #---------------------------------------------------------------------
    # Schedule U1 before U2 when U1 == WR(x) < RD(x) == U2: combinational
    #
    # Explicitly, one should define these to invert the implicit constraint:
    # - RD(x) < U when U == WR(x) --> RD(x) ( == U') < U == WR(x)
    # - WR(x) > U when U == RD(x) --> RD(x) == U < WR(x) ( == U')
    # constraint RD(x) < U1 & U2 reads  x --> U2 == RD(x) <  U1
    # constraint RD(x) > U1 & U2 reads  x --> U1 <  RD(x) == U2 # impl
    # constraint WR(x) < U1 & U2 writes x --> U2 == WR(x) <  U1 # impl
    # constraint WR(x) > U1 & U2 writes x --> U1 <  WR(x) == U2
    # Doesn't work for nested data struct and slice:

    read_upblks = defaultdict(set)
    write_upblks = defaultdict(set)

    constraint_objs = defaultdict(set)
    expl_constraints = set( top._dsl.all_U_U_constraints )

    for data in [ upblk_reads, genblk_reads ]:
      for blk, reads in data.items():
        for rd in reads:
          read_upblks[ rd ].add( blk )

    for data in [ upblk_writes, genblk_writes ]:
      for blk, writes in data.items():
        for wr in writes:
          write_upblks[ wr ].add( blk )

    for typ in [ 'rd', 'wr' ]: # deduplicate code
      if typ == 'rd':
        constraints = RD_U
        equal_blks  = read_upblks
      else:
        constraints = WR_U
        equal_blks  = write_upblks

      # enumerate variable objects
      for obj, constrained_blks in constraints.items():

        # enumerate upblks that has a constraint with x
        for (sign, co_blk) in constrained_blks:

          for eq_blk in equal_blks[ obj ]: # blocks that are U == RD(x)
            if co_blk != eq_blk:
              if sign == 1: # RD/WR(x) < U is 1, RD/WR(x) > U is -1
                # eq_blk == RD/WR(x) < co_blk
                U_U.add( (eq_blk, co_blk) )
                constraint_objs[ (eq_blk, co_blk) ].add( obj )
              else:
                # co_blk < RD/WR(x) == eq_blk
                U_U.add( (co_blk, eq_blk) )
                constraint_objs[ (co_blk, eq_blk) ].add( obj )

    #---------------------------------------------------------------------
    # Implicit constraint
    #---------------------------------------------------------------------
    # Synthesize total constraints between two upblks that read/write to
    # the "same variable" (we also handle the read/write of a recursively
    # nested field/slice)
    #
    # Implicitly, WR(x) < RD(x), so when U1 writes X and U2 reads x
    # - U1 == WR(x) & U2 == RD(x) --> U1 == WR(x) < RD(x) == U2

    impl_constraints = set()

    # Collect all objs that write the variable whose id is "read"
    # 1) RD A.b.b     - WR A.b.b, A.b, A
    # 2) RD A.b[1:10] - WR A.b[1:10], A.b, A
    # 3) RD A.b[1:10] - WR A.b[0:5], A.b[6], A.b[8:11]

    for obj, rd_blks in read_upblks.items():
      writers = []

      # Check parents. Cover 1) and 2)
      x = obj
      while x.is_signal():
        if x in write_upblks:
          writers.append( x )
        x = x.get_parent_object()

      # Check the sibling slices. Cover 3)
      if obj.is_signal():
        for x in obj.get_sibling_slices():
          if x.slice_overlap( obj ) and x in write_upblks:
            writers.append( x )

      # Add all constraints
      for writer in writers:
        for wr_blk in write_upblks[ writer ]:
          if wr_blk not in update_ff:
            for rd_blk in rd_blks:
              if wr_blk != rd_blk:
                # if rd_blk not in update_ff:
                impl_constraints.add( (wr_blk, rd_blk) ) # wr < rd default
                constraint_objs[ (wr_blk, rd_blk) ].add( obj )

    # Collect all objs that read the variable whose id is "write"
    # 1) WR A.b.b.b, A.b.b, A.b, A (detect 2-writer conflict)
    # 2) WR A.b.b.b   - RD A.b.b, A.b, A
    # 3) WR A.b[1:10] - RD A.b[1:10], A,b, A
    # 4) WR A.b[1:10], A.b[0:5], A.b[6] (detect 2-writer conflict)
    # "WR A.b[1:10] - RD A.b[0:5], A.b[6], A.b[8:11]" has been discovered

    for obj, wr_blks in write_upblks.items():
      readers = []

      # Check parents. Cover 2) and 3). 1) and 4) should be detected in elaboration
      x = obj
      while x.is_signal():
        if x in read_upblks:
          readers.append( x )
        x = x.get_parent_object()

      # Add all constraints
      for wr_blk in wr_blks:
        if wr_blk not in update_ff:
          for reader in readers:
              for rd_blk in read_upblks[ reader ]:
                if wr_blk != rd_blk:
                  # if rd_blk not in update_ff:
                  impl_constraints.add( (wr_blk, rd_blk) ) # wr < rd default
                  constraint_objs[ (wr_blk, rd_blk) ].add( obj )

    top._dag.constraint_objs = constraint_objs
    top._dag.all_constraints = { *U_U }
    for (x, y) in impl_constraints:
      if (y, x) not in U_U: # no conflicting expl
        top._dag.all_constraints.add( (x, y) )

  #-----------------------------------------------------------------------
  # Process methods
  #----------------------------------------------------------------------
  # I assume method don't call other methods here

  # Do bfs to find out all potential total constraints associated with
  # each method, direction conflicts, and incomplete constraints

  def _process_methods( self, top ):
    _, _, _, all_M_constraints = top.get_all_explicit_constraints()

    # Here we collect all top level callee ports and collect all
    # constraints that involves a top level callee. NOTE THAT it is
    # possible that the top level callee is connected to an actual callee
    # somewhere else. Hence we use the ACTUAL METHOD as identifier
    # because all members in the net will eventually point to the same
    # method object.

    top._dsl.top_level_callee_ports = top.get_all_object_filter(
      lambda x: isinstance(x, CalleePort) and x.get_host_component() is top )

    method_is_top_level_callee = set()

    all_method_nets = top.get_all_method_nets()
    for writer, net in all_method_nets:
      if writer is not None:
        for member in net:
          if member is not writer:
            assert member.method is None
            member.method = writer.method

          # If the member is a top level callee, we add the writer's
          # actual method to the set
          if member.get_host_component() is top:
            # if not writer.method.__name__.startswith('_binded_method'):
            method_is_top_level_callee.add( writer.method )

    # Add those callee ports that are not part of a net
    for callee in top._dsl.top_level_callee_ports:
      if callee.method:
        # if not callee.method.__name__.startswith('_binded_method'):
        method_is_top_level_callee.add( callee.method )

    method_blks = defaultdict(set)

    # Collect each CalleePort/method is called in which update block
    # We use the actual method of CalleePort to identify each call
    for blk, calls in top._dsl.all_upblk_calls.items():
      for call in calls:
        if isinstance( call, MethodPort ):
          method_blks[ call.method ].add( blk )
        elif isinstance( call, (NonBlockingIfc, BlockingIfc) ):
          method_blks[ call.method.method ].add( blk )
        else:
          method_blks[ call ].add( blk )

    # Put all M-related constraints into predecessor and successor dicts
    pred = defaultdict(set)
    succ = defaultdict(set)

    top._dag.top_level_callee_constraints = set()

    # We pre-process M(x) == M(y) constraints into per-method equivalence
    # sets. We have to do it here for potential open-loop constraints

    equiv = defaultdict(set)
    for (x, y, is_equal) in all_M_constraints:

      if is_equal: # M(x) == M(y)
        # Use the actual method object for constraints

        if   isinstance( x, MethodPort ):
          xx = x.method
        elif isinstance( x, (NonBlockingIfc, BlockingIfc) ):
          xx = x.method.method
        else:
          xx = x

        if   isinstance( y, MethodPort ):
          yy = y.method
        elif isinstance( y, (NonBlockingIfc, BlockingIfc) ):
          yy = y.method.method
        else:
          yy = y

        equiv[xx].add( yy )
        equiv[yy].add( xx )

    # flood-fill to find out all equivalent classes

    visited = set()
    for x in equiv:
      if x not in visited:
        equiv_class = set()
        visited.add( x )
        Q = deque( [x] )
        while Q:
          u = Q.popleft()
          equiv_class.add(u)

          for v in equiv[u]:
            if v not in visited:
              visited.add(v)
              Q.append(v)

        # Point all nodes in the equivalence class to the same set
        for u in equiv_class:
          equiv[ u ] = equiv_class

    for (x, y, is_equal) in all_M_constraints:
      if is_equal: continue

      # Use the actual method object for constraints

      if   isinstance( x, MethodPort ):
        xx = x.method
      elif isinstance( x, (NonBlockingIfc, BlockingIfc) ):
        xx = x.method.method
      else:
        xx = x

      if   isinstance( y, MethodPort ):
        yy = y.method
      elif isinstance( y, (NonBlockingIfc, BlockingIfc) ):
        yy = y.method.method
      else:
        yy = y

      pred[ yy ].add( xx )
      succ[ xx ].add( yy )

      if xx in equiv: # xx is in a equivalence class
        for zz in equiv[xx]:
          if zz in method_is_top_level_callee:
            top._dag.top_level_callee_constraints.add( (zz, yy) )
      else:
        if xx in method_is_top_level_callee:
          top._dag.top_level_callee_constraints.add( (xx, yy) )

      if yy in equiv: # yy is in a equivalence class
        for zz in equiv[yy]:
          if zz in method_is_top_level_callee:
            top._dag.top_level_callee_constraints.add( (xx, zz) )

      else:
        if yy in method_is_top_level_callee:
          top._dag.top_level_callee_constraints.add( (xx, yy) )

    verbose = False

    all_upblks = top.get_all_update_blocks()

    for method, assoc_blks in method_blks.items():
      visited = {  (method, 0)  }
      Q = deque( [ (method, 0) ] ) # -1: pred, 0: don't know, 1: succ

      if verbose: print()
      while Q:
        (u, w) = Q.pop()
        if verbose: print((u, w))

        if u in equiv:
          for v in equiv[u]:
            if (v, w) not in visited:
              visited.add( (v, w) )
              Q.append( (v, w) )

        if w <= 0:
          for v in pred[u]:

            if v in all_upblks:
              # Find total constraint (v < blk) by v < method_u < method_u'=blk
              # INVALID if we have explicit constraint (blk < method_u)

              for blk in assoc_blks:
                if blk not in pred[u]:
                  if v != blk:
                    if verbose: print("w<=0, v is blk".center(10),v, blk)
                    if verbose: print(v.__name__.center(25)," < ", \
                                blk.__name__.center(25))
                    top._dag.all_constraints.add( (v, blk) )

            else:
              if v in method_blks:
                # TODO Now I'm leaving incomplete dependency chain because I didn't close the circuit loop.
                # E.g. I do port.wr() somewhere in __main__ to write to a port.

                # Find total constraint (vb < blk) by vb=method_v < method_u=blk
                # INVALID if we have explicit constraint (blk < method_v) or (method_u < vb)

                v_blks = method_blks[ v ]
                for vb in v_blks:
                  if vb not in succ[u]:
                    for blk in assoc_blks:
                      if blk not in pred[v]:
                        if vb != blk:
                          if verbose: print("w<=0, v is method".center(10),v, blk)
                          if verbose: print(vb.__name__.center(25)," < ", \
                                      blk.__name__.center(25))
                          top._dag.all_constraints.add( (vb, blk) )

              if (v, -1) not in visited:
                visited.add( (v, -1) )
                Q.append( (v, -1) ) # ? < v < u < ... < method < blk_id

        if w >= 0:
          for v in succ[u]:

            if v in all_upblks:
              # Find total constraint (blk < v) by blk=method_u' < method_u < v
              # INVALID if we have explicit constraint (method_u < blk)

              for blk in assoc_blks:
                if blk not in succ[u]:
                  if v != blk:
                    if verbose: print("w>=0, v is blk".center(10),blk, v)
                    if verbose: print(blk.__name__.center(25)," < ", \
                                      v.__name__.center(25))
                    top._dag.all_constraints.add( (blk, v) )

            else:
              if v in method_blks:
                # assert v in method_blks, "Incomplete elaboration, something is wrong! %s" % hex(v)
                # TODO Now I'm leaving incomplete dependency chain because I didn't close the circuit loop.
                # E.g. I do port.wr() somewhere in __main__ to write to a port.

                # Find total constraint (blk < vb) by blk=method_u < method_v=vb
                # INVALID if we have explicit constraint (vb < method_u) or (method_v < blk)

                v_blks = method_blks[ v ]
                for vb in v_blks:
                  if not vb in pred[u]:
                    for blk in assoc_blks:
                      if not blk in succ[v]:
                        if vb != blk:
                          if verbose: print("w>=0, v is method".center(10), blk, v)
                          if verbose: print(blk.__name__.center(25)," < ", \
                                            vb.__name__.center(25))
                          top._dag.all_constraints.add( (blk, vb) )

              if (v, 1) not in visited:
                visited.add( (v, 1) )
                Q.append( (v, 1) ) # blk_id < method < ... < u < v < ?

    # Mark update blocks that call blocking methods
    # (CalleeIfcFL/CallerIfcFL) for greenlet wrapping

    blocking_ifcs = top.get_all_object_filter( lambda x: isinstance( x, (CalleeIfcFL, CallerIfcFL) ) )

    top._dag.greenlet_upblks = set()

    for blocking_method in blocking_ifcs:
      for blk in method_blks[ blocking_method.method.method ]:
        top._dag.greenlet_upblks.add( blk )



"""
========================================================================
BasePass.py
========================================================================
An "abstract" base class for all passes.

Author : Shunning Jiang
Date   : Dec 17, 2017
"""
class PassMetadata:
  pass

class BasePass:
  def __init__( self, debug=False ): # initialize parameters
    self.debug = debug

  def __call__( self, m ): # execute pass on model m
    pass



#=========================================================================
# BehavioralRTLIRL2Pass_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Feb 2, 2019
"""Test the level 2 behavioral RTLIR passes.

The L2 generation, L2 type check, and visualization passes are invoked. The
generation pass results are verified against a reference AST.
"""

from pymtl3.passes.rtlir.behavioral.BehavioralRTLIR import *
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRGenL2Pass import (
    BehavioralRTLIRGenL2Pass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRTypeCheckL2Pass import (
    BehavioralRTLIRTypeCheckL2Pass,
)
from pymtl3.passes.rtlir.errors import PyMTLSyntaxError, PyMTLTypeError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure
from pymtl3.passes.testcases import (
    CaseAddComponentComp,
    CaseAddStructBits1Comp,
    CaseAndStructComp,
    CaseComponentEndRangeComp,
    CaseComponentIfCondComp,
    CaseComponentIfExpBodyComp,
    CaseComponentIfExpCondComp,
    CaseComponentIfExpElseComp,
    CaseComponentStartRangeComp,
    CaseComponentStepRangeComp,
    CaseDifferentTypesIfExpComp,
    CaseExplicitBoolComp,
    CaseForLoopElseComp,
    CaseFuncCallAfterInComp,
    CaseInvalidBreakComp,
    CaseInvalidContinueComp,
    CaseInvalidDivComp,
    CaseInvalidInComp,
    CaseInvalidIsComp,
    CaseInvalidIsNotComp,
    CaseInvalidNotInComp,
    CaseInvComponentComp,
    CaseLambdaConnectComp,
    CaseLambdaConnectWithListComp,
    CaseMultiOpComparisonComp,
    CaseNoArgsToRangeComp,
    CaseNotStructComp,
    CaseRedefLoopIndexComp,
    CaseSignalAfterInComp,
    CaseSignalAsLoopIndexComp,
    CaseStructIfCondComp,
    CaseStructIfExpCondComp,
    CaseTmpVarUsedBeforeAssignmentComp,
    CaseTooManyArgsToRangeComp,
    CaseVariableStepRangeComp,
    CaseZeroStepRangeComp,
)


def local_do_test( m ):
  """Check if generated behavioral RTLIR is the same as reference."""
  if isinstance(m, type):
    m = m.DUT()
  m.elaborate()
  m.apply( BehavioralRTLIRGenL2Pass() )
  m.apply( BehavioralRTLIRTypeCheckL2Pass() )

  try:
    ref = m._rtlir_test_ref
    for blk in m.get_update_blocks():
      upblk = m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ]
      assert upblk == ref[ blk.__name__ ]
  except AttributeError:
    pass

#-------------------------------------------------------------------------
# Correct test cases
#-------------------------------------------------------------------------

def test_L2_lambda_connect( do_test ):
  a = CaseLambdaConnectComp.DUT()
  a._rtlir_test_ref = { '_lambda__s_out' : CombUpblk( '_lambda__s_out', [
    Assign( Attribute( Base( a ), 'out' ),
            BinOp(Attribute(Base(a), 'in_'), Add(), SizeCast(32, Number(42))), True
        ) ] ) }
  do_test( a )

def test_L2_lambda_connect_with_list( do_test ):
  a = CaseLambdaConnectWithListComp.DUT()
  a._rtlir_test_ref = { '_lambda__s_out_1_' : CombUpblk( '_lambda__s_out_1_', [
    Assign( Index( Attribute( Base( a ), 'out' ), Number(1) ),
            BinOp(Attribute(Base(a), 'in_'), Add(), SizeCast(32, Number(42))), True
        ) ] ) }
  do_test( a )

#-------------------------------------------------------------------------
# PyMTL type errors
#-------------------------------------------------------------------------

def test_L2_add_component( do_test ):
  with expected_failure( PyMTLTypeError, "rhs of binop should be signal/const" ):
    do_test( CaseAddComponentComp )

def test_L2_inv_component( do_test ):
  with expected_failure( PyMTLTypeError, "only applies to signals and consts" ):
    do_test( CaseInvComponentComp )

def test_L2_for_start_component( do_test ):
  with expected_failure( PyMTLTypeError, "must be a constant expression" ):
    do_test( CaseComponentStartRangeComp )

def test_L2_for_end_component( do_test ):
  with expected_failure( PyMTLTypeError, "must be a constant expression" ):
    do_test( CaseComponentEndRangeComp )

def test_L2_for_step_component( do_test ):
  with expected_failure( PyMTLTypeError, "must be a constant expression" ):
    do_test( CaseComponentStepRangeComp )

def test_L2_if_cond_component( do_test ):
  with expected_failure( PyMTLTypeError, "condition of if must be a signal" ):
    do_test( CaseComponentIfCondComp )

def test_L2_ifexp_cond_component( do_test ):
  with expected_failure( PyMTLTypeError, "condition of if-exp must be a signal" ):
    do_test( CaseComponentIfExpCondComp )

def test_L2_ifexp_body_component( do_test ):
  with expected_failure( PyMTLTypeError, "body of if-exp must be a signal" ):
    do_test( CaseComponentIfExpBodyComp )

def test_L2_ifexp_orelse_component( do_test ):
  with expected_failure( PyMTLTypeError, "else branch of if-exp must be a signal" ):
    do_test( CaseComponentIfExpElseComp )

def test_L2_if_cond_bool( do_test ):
  with expected_failure( PyMTLTypeError, "cannot be converted to bool" ):
    do_test( CaseStructIfCondComp )

def test_L2_for_step_zero( do_test ):
  with expected_failure( PyMTLTypeError, "step of for-loop cannot be zero" ):
    do_test( CaseZeroStepRangeComp )

def test_L2_for_step_variable( do_test ):
  with expected_failure( PyMTLTypeError, "step of a for-loop must be a constant" ):
    do_test( CaseVariableStepRangeComp )

def test_L2_ifexp_cond_bool( do_test ):
  with expected_failure( PyMTLTypeError, "cannot be converted to bool" ):
    do_test( CaseStructIfExpCondComp )

def test_L2_ifexp_body_else_diff_type( do_test ):
  with expected_failure( PyMTLTypeError, "must have the same type" ):
    do_test( CaseDifferentTypesIfExpComp )

def test_L2_unary_not_cast_to_bool( do_test ):
  with expected_failure( PyMTLTypeError, "cannot be cast to bool" ):
    do_test( CaseNotStructComp )

def test_L2_bool_cast_to_bool( do_test ):
  with expected_failure( PyMTLTypeError, "cannot be cast into bool" ):
    do_test( CaseAndStructComp )

def test_L2_binop_non_vector( do_test ):
  with expected_failure( PyMTLTypeError, "should be of vector type" ):
    do_test( CaseAddStructBits1Comp )

#-------------------------------------------------------------------------
# PyMTL syntax errors
#-------------------------------------------------------------------------

def test_L2_call_bool( do_test ):
  with expected_failure( PyMTLSyntaxError, "bool values cannot be instantiated explicitly" ):
    do_test( CaseExplicitBoolComp )

def test_L2_tmp_var_used_before_assignement( do_test ):
  with expected_failure( PyMTLSyntaxError, "tmpvar u used before assignment" ):
    do_test( CaseTmpVarUsedBeforeAssignmentComp )

def test_L2_for_else( do_test ):
  with expected_failure( PyMTLSyntaxError, "for loops cannot have 'else' branch" ):
    do_test( CaseForLoopElseComp )

def test_L2_for_index_signal( do_test ):
  with expected_failure( PyMTLSyntaxError, "loop index must be a temporary variable" ):
    do_test( CaseSignalAsLoopIndexComp )

def test_L2_for_index_redefined( do_test ):
  with expected_failure( PyMTLSyntaxError, "Redefinition of loop index i" ):
    do_test( CaseRedefLoopIndexComp )

def test_L2_for_iter_name( do_test ):
  with expected_failure( PyMTLSyntaxError, "only use range() after 'in'" ):
    do_test( CaseSignalAfterInComp )

def test_L2_for_iter_call_non_range( do_test ):
  with expected_failure( PyMTLSyntaxError, "only use range() after 'in'" ):
    do_test( CaseFuncCallAfterInComp )

def test_L2_for_iter_range_arg_0( do_test ):
  with expected_failure( PyMTLSyntaxError, "1~3 arguments should be given to range" ):
    do_test( CaseNoArgsToRangeComp )

def test_L2_for_iter_range_arg_4( do_test ):
  with expected_failure( PyMTLSyntaxError, "1~3 arguments should be given to range" ):
    do_test( CaseTooManyArgsToRangeComp )

def test_L2_invalid_op_is( do_test ):
  with expected_failure( PyMTLSyntaxError, "not supported" ):
    do_test( CaseInvalidIsComp )

def test_L2_invalid_op_is_not( do_test ):
  with expected_failure( PyMTLSyntaxError, "not supported" ):
    do_test( CaseInvalidIsNotComp )

def test_L2_invalid_op_in( do_test ):
  with expected_failure( PyMTLSyntaxError, "not supported" ):
    do_test( CaseInvalidInComp )

def test_L2_invalid_op_not_in( do_test ):
  with expected_failure( PyMTLSyntaxError, "not supported" ):
    do_test( CaseInvalidNotInComp )

def test_L2_invalid_op_floor_div( do_test ):
  with expected_failure( PyMTLSyntaxError, "not supported" ):
    do_test( CaseInvalidDivComp )

def test_L2_compare_terms( do_test ):
  with expected_failure( PyMTLSyntaxError, "Comparison can only have 2 operands" ):
    do_test( CaseMultiOpComparisonComp )

#-------------------------------------------------------------------------
# PyMTL syntax errors for unsupported Python syntax
#-------------------------------------------------------------------------

def test_Break( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: break" ):
    do_test( CaseInvalidBreakComp )

def test_Continue( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: continue" ):
    do_test( CaseInvalidContinueComp )



from pymtl3 import *


class RegisterFile( Component ):

  def construct( s, Type, nregs=32, rd_ports=1, wr_ports=1,
                const_zero=False ):

    addr_type = mk_bits( clog2( nregs ) )

    s.raddr = [ InPort( addr_type ) for i in range( rd_ports ) ]
    s.rdata = [ OutPort( Type ) for i in range( rd_ports ) ]

    s.waddr = [ InPort( addr_type ) for i in range( wr_ports ) ]
    s.wdata = [ InPort( Type ) for i in range( wr_ports ) ]
    s.wen   = [ InPort( Bits1 ) for i in range( wr_ports ) ]

    s.regs = [ Wire( Type ) for i in range(nregs) ]

    @s.update
    def up_rf_read():
      for i in range( rd_ports ):
        s.rdata[i] = s.regs[ s.raddr[i] ]

    if const_zero:
      @s.update_ff
      def up_rf_write_constzero():
        for i in range( wr_ports ):
          if s.wen[i] & (s.waddr[i] != addr_type(0)):
            s.regs[ s.waddr[i] ] <<= s.wdata[i]
    else:
      @s.update_ff
      def up_rf_write():
        for i in range( wr_ports ):
          if s.wen[i]:
            s.regs[ s.waddr[i] ] <<= s.wdata[i]



#=========================================================================
# PrintWavePass_test.py
#=========================================================================
# A simple demo of how PrintWavePass is used in the simulation of
# a toy component.
#
# Author : Kaishuo Cheng
# Date   : Oct 8th, 2019

import io
from contextlib import redirect_stdout

from pymtl3.datatypes import (
    Bits1,
    Bits16,
    Bits32,
    Bits128,
    b1,
    b16,
    b32,
    b128,
    bitstruct,
)
from pymtl3.dsl import *
from pymtl3.passes import TracingConfigs
from pymtl3.passes.errors import ModelTypeError
from pymtl3.passes.PassGroups import SimulationPass


def test_toy():
  class Toy( Component ):
    """Toy adder component"""

    def construct( s ):
      # Interfaces
      s.i = InPort( Bits32 )
      s.inlong = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.state = Wire(Bits1)
      @s.update
      def add_upblk():
        # This update block models the behavior of a 32-bit adder
        s.out = s.i + s.inlong
        if s.out[3] == "1":
          s.state = s.state +b1(1)
        else:
          s.state = b1(0)

  # Create a toy component and elaborate it
  dut = Toy()
  dut.config_tracing = TracingConfigs( tracing='text_ascii' )

  # Setup the simulation
  dut.apply( SimulationPass() )
  dut.sim_reset()
  # Test vector
  vector = [
    #  i        inlong       out
    b32(1),    b32(2),    b32(3),
    b32(0),    b32(2),    b32(2),
    b32(0),    b32(2),    b32(2),
    b32(1),    b32(-2),   b32(-1),
    b32(1),    b32(-42),  b32(-41),
    b32(1),    b32(-4),   b32(-3),
    b32(1),    b32(2),    b32(3),
    b32(0),    b32(2),    b32(2),
    b32(1),    b32(2),    b32(3),
    b32(0),    b32(-5),   b32(-5),
  ]

  # Begin simulation
  for i, inlong, out in zip(vector[0::3], vector[1::3], vector[2::3]):
    dut.i = i
    dut.inlong = inlong
    dut.eval_combinational()
    dut.tick()
    assert dut.out == out

  #print
  f = io.StringIO()
  dut.print_textwave()
  with redirect_stdout(f):
    dut.print_textwave()
  out = f.getvalue()
  for i in dut._tracing.text_sigs:
    dot = i.find(".")
    sliced = i[dot+1:]
    if sliced != "reset" and sliced != "clk":
      assert i[dot+1:] in out

def test_widetoy():
  class Toy( Component ):

    def construct( s ):
      # Interfaces
      s.i = InPort( Bits128 )
      s.inlong = InPort( Bits128 )
      s.out = OutPort( Bits128 )
      s.state = Wire(Bits1)
      @s.update
      def add_upblk():
        # This update block models the behavior of a 32-bit adder
        s.out = s.i + s.inlong
        if s.out[3] == "1":
          s.state = s.state +b1(1)
        else:
          s.state = b1(0)

  # Create a toy component and elaborate it
  dut = Toy()

  dut.config_tracing = TracingConfigs( tracing='text_fancy' )

  dut.elaborate()

  # Setup the simulation
  dut.apply( SimulationPass() )
  dut.sim_reset()
  # Test vector
  vector = [
    #  i        inlong       out
    b128(0x1000000000000+1),    b128(2),    b128(0x1000000000000+3),
    b128(0x1000000000000+0),    b128(2),    b128(0x1000000000000+2),
    b128(0x1000000000000+0),    b128(2),    b128(0x1000000000000+2),
    b128(0x1000000000000+1),    b128(-2),   b128(0x1000000000000+-1),
    b128(0x1000000000000+1),    b128(-42),  b128(0x1000000000000+-41),
    b128(0x1000000000000+1),    b128(-4),   b128(0x1000000000000+-3),
    b128(0x1000000000000+1),    b128(2),    b128(0x1000000000000+3),
    b128(0x1000000000000+0),    b128(2),    b128(0x1000000000000+2),
    b128(0x1000000000000+1),    b128(2),    b128(0x1000000000000+3),
    b128(0x1000000000000+0),    b128(-5),   b128(0x1000000000000+-5),
  ]

  # Begin simulation
  for i, inlong, out in zip(vector[0::3], vector[1::3], vector[2::3]):
    dut.i = i
    dut.inlong = inlong
    dut.eval_combinational()
    dut.tick()
    assert dut.out == out

  #print
  f = io.StringIO()
  dut.print_textwave()
  with redirect_stdout(f):
    dut.print_textwave()
  out = f.getvalue()
  for i in dut._tracing.text_sigs:
    dot = i.find(".")
    sliced = i[dot+1:]
    if sliced != "reset" and sliced != "clk":
      assert i[dot+1:] in out

def test_bitstruct():

  @bitstruct
  class XX:
    x: Bits16
    y: Bits16

  class Toy( Component ):
    def construct( s ):
      # Interfaces
      s.i = InPort( XX )
      s.inlong = InPort( Bits16 )
      s.out = OutPort( XX )
      s.state = Wire(Bits1)
      @s.update
      def add_upblk():
        # This update block models the behavior of a 32-bit adder
        s.out.x = s.i.x + s.inlong
        s.out.y = s.i.y + s.inlong
        if s.out.x[3] == "1":
          s.state = s.state +b1(1)
        else:
          s.state = b1(0)

  # Create a toy component and elaborate it
  dut = Toy()

  dut.config_tracing = TracingConfigs( tracing='text_fancy' )

  dut.elaborate()

  # Setup the simulation
  dut.apply( SimulationPass() )

  dut.sim_reset()
  # Test vector
  vector = [
    #  i        inlong       out
    XX(1,1),    b16(2),    XX(3,3),
    XX(0,0),    b16(2),    XX(2,2),
    XX(0,0),    b16(2),    XX(2,2),
    XX(1,1),    b16(-2),   XX(-1,-1),
    XX(1,1),    b16(-42),  XX(-41,-41),
    XX(1,1),    b16(-4),   XX(-3,-3),
    XX(1,1),    b16(2),    XX(3,3),
    XX(0,0),    b16(2),    XX(2,2),
    XX(1,1),    b16(2),    XX(3,3),
    XX(0,0),    b16(-5),   XX(-5,-5),
  ]

  # Begin simulation
  for i, inlong, out in zip(vector[0::3], vector[1::3], vector[2::3]):
    dut.i = i
    dut.inlong = inlong
    dut.eval_combinational()
    dut.tick()
    assert dut.out == out

  #print
  f = io.StringIO()
  dut.print_textwave()
  with redirect_stdout(f):
    dut.print_textwave()
  out = f.getvalue()
  for i in dut._tracing.text_sigs:
    dot = i.find(".")
    sliced = i[dot+1:]
    if sliced != "reset" and sliced != "clk":
      assert i[dot+1:] in out



"""
========================================================================
SimpleSchedulePass.py
========================================================================
Generate a simple schedule (no Mamba techniques here) based on the
DAG generated by some previous pass.

Author : Shunning Jiang
Date   : Dec 26, 2018
"""

from collections import defaultdict

from pymtl3.dsl.errors import UpblkCyclicError
from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.errors import PassOrderError


class SimpleSchedulePass( BasePass ):
  def __call__( self, top ):
    if not hasattr( top._dag, "all_constraints" ):
      raise PassOrderError( "all_constraints" )

    top._sched = PassMetadata()

    self.schedule_intra_cycle( top )
    self.schedule_ff( top )
    self.schedule_posedge_flip( top )

  def schedule_intra_cycle( self, top ):

    if not hasattr( top, "_sched" ):
      raise Exception( "Please create top._sched pass metadata namespace first!" )

    # Construct the intra-cycle graph based on normal update blocks

    V   = top._dag.final_upblks - top.get_all_update_ff()
    E   = set()
    Es  = { v: [] for v in V }
    InD = { v: 0  for v in V }

    for (u, v) in top._dag.all_constraints: # u -> v
      if u in V and v in V:
        InD[v] += 1
        Es[u].append( v )
        E.add( (u, v) )

    import os
    if 'MAMBA_DAG' in os.environ:
      dump_dag( top, V, E )

    # Perform topological sort for a serial schedule.

    top._sched.update_schedule = update_schedule = []

    Q = [ v for v in V if not InD[v] ]

    import random
    while Q:
      random.shuffle(Q)
      u = Q.pop()
      update_schedule.append( u )
      for v in Es[u]:
        InD[v] -= 1
        if not InD[v]:
          Q.append( v )

    check_schedule( top, update_schedule, V, E, InD )

  def schedule_ff( self, top ):

    if not hasattr( top, "_sched" ):
      raise Exception( "Please create top._sched pass metadata namespace first!" )
    top._sched.schedule_ff = list( top.get_all_update_ff().copy() )

  def schedule_posedge_flip( self, top ):

    if not hasattr( top, "_sched" ):
      raise Exception( "Please create top._sched pass metadata namespace first!" )

    # To reduce the time to compile the code and the amount of bytecode, I
    # use a heuristic to group signals that belong to
    #   s.x.y.z._flip()
    #   s.x.y.zz._flip()
    # becomes
    #   x = s.x.y
    #   x.z._flip()
    #   x.zz._flip()

    hostobj_signals = defaultdict(list)
    for x in reversed(sorted( top._dsl.all_signals, \
        key=lambda x: x.get_host_component().get_component_level() )):
      if x._dsl.needs_double_buffer:
        hostobj_signals[ x.get_host_component() ].append( x )

    done = False
    while not done:
      next_hostobj_signals = defaultdict(list)
      done = True

      for x, y in hostobj_signals.items():
        if len(y) > 1:
          next_hostobj_signals[x].extend( y )
        elif x is top:
          next_hostobj_signals[x].extend( y )
        else:
          x = x.get_parent_object()
          next_hostobj_signals[x].append( y[0] )
          done = False
      hostobj_signals = next_hostobj_signals

    strs = []
    for x,y in hostobj_signals.items():
      if len(y) == 1:
        strs.append( f"{repr(y[0])}._flip()" )
      elif x is top:
        for z in sorted(y, key=repr):
          strs.append(f"{repr(z)}._flip()")
      else:
        pos = len(repr(x)) + 1
        strs.append( f"x = {repr(x)}" )

        for z in sorted(y, key=repr):
          strs.append(f"x.{repr(z)[pos:]}._flip()")

    if not strs:
      def no_double_buffer():
        pass
      top._sched.schedule_posedge_flip = [ no_double_buffer ]

    else:
      src = """
      def compile_double_buffer( s ):
        def double_buffer():
          {}
        return double_buffer
      """.format( "\n          ".join(strs) )

      import py
      # print(src)
      l = locals()
      exec(py.code.Source( src ).compile(), l)

      top._sched.schedule_posedge_flip = [ l['compile_double_buffer']( top ) ]

def dump_dag( top, V, E ):
  from graphviz import Digraph
  from pymtl3.dsl import CalleePort
  dot = Digraph()
  dot.graph_attr["rank"] = "same"
  dot.graph_attr["ratio"] = "compress"
  dot.graph_attr["margin"] = "0.1"

  for x in V:
    x_name = repr(x) if isinstance( x, CalleePort ) else x.__name__
    if x in top._dsl.all_update_ff:
      x_name += "_FF"
    try:
      x_host = repr(x.get_parent_object() if isinstance( x, CalleePort )
                    else top.get_update_block_host_component(x))
    except:
      x_host = ""
    dot.node( x_name +"\\n@" + x_host, shape="box")

  for (x, y) in E:
    x_name = repr(x) if isinstance( x, CalleePort ) else x.__name__
    if x in top._dsl.all_update_ff:
      x_name += "_FF"
    try:
      x_host = repr(x.get_parent_object() if isinstance( x, CalleePort )
                    else top.get_update_block_host_component(x))
    except:
      x_host = ""
    y_name = repr(y) if isinstance( y, CalleePort ) else y.__name__
    if y in top._dsl.all_update_ff:
      y_name += "_FF"
    try:
      y_host = repr(y.get_parent_object() if isinstance( y, CalleePort )
                    else top.get_update_block_host_component(y))
    except:
      y_host = ""

    dot.edge( x_name+"\\n@"+x_host, y_name+"\\n@"+y_host )
  dot.render( "/tmp/upblk-dag.gv", view=True )

def check_schedule( top, schedule, V, E, in_degree ):

  if len(schedule) != len(V):
    V_leftovers = {  v for v in V if in_degree[v]  }
    E_leftovers = {  (x,y) for (x,y) in E
                         if x in V_leftovers and y in V_leftovers  }
    dump_dag( top, V_leftovers, E_leftovers )

    raise UpblkCyclicError( """
Update blocks have cyclic dependencies.
* Please consult update dependency graph for details.
    """)



"""
#=========================================================================
# MemIfcs
#=========================================================================
#
# Author: Shunning Jiang
# Date  : May 19, 2019
"""
from pymtl3 import *
from pymtl3.stdlib.ifcs import MemMsgType, mk_mem_msg

from .mem_ifcs import MemMasterIfcFL, MemMinionIfcCL


def test_mem_fl_cl_adapter():

  class SomeMasterFL( Component ):
    def construct( s, base, has_loop=0 ):
      s.mem = MemMasterIfcFL()

      s.addr = 0x1000 + base
      s.end  = 0x1000 + base + 0x10

      s.trace = "                   "

      if has_loop == 1:
        @s.update
        def up_master_while():
          s.trace = "                  "
          while s.addr < s.end:
            s.mem.write( s.addr, 4, 0xdead0000 | s.addr )
            s.trace = "wr 0x{:x}         ".format( s.addr )
            x = s.mem.read( s.addr, 4 )
            s.trace = "rd 0x{:x} {}".format( s.addr, x )
            s.addr += 4
      else:
        @s.update
        def up_master_noloop():
          s.trace = "#                 "
          s.mem.write( s.addr, 4, 0xdead0000 | s.addr )
          s.trace = "wr 0x{:x}         ".format( s.addr )
          x = s.mem.read( s.addr, 4 )
          s.trace = "rd 0x{:x} {}".format( s.addr, x )
          s.addr += 4

    def done( s ):
      return s.addr >= s.end

    def line_trace( s ):
      return s.trace

  class SomeMinionCL( Component ):

    def recv( s, msg ):
      assert s.entry is None
      s.entry = msg

    def recv_rdy( s ):
      return s.entry is None

    def read( s, addr, nbytes ):
      nbytes = int(nbytes)
      nbits = nbytes << 3
      ret, shamt = Bits( nbits, 0 ), Bits( nbits, 0 )
      addr = int(addr)
      end  = addr + nbytes
      for j in range( addr, end ):
        ret += Bits( nbits, s.memory[j] ) << shamt
        shamt += Bits4(8)
      return ret

    def write( s, addr, nbytes, data ):
      tmp  = int(data)
      addr = int(addr)
      end  = addr + int(nbytes)
      for j in range( addr, end ):
        s.memory[j] = tmp & 255
        tmp >>= 8

    def construct( s, req_class, resp_class ):
      s.mem = MemMinionIfcCL( req_class, resp_class, s.recv, s.recv_rdy )
      s.entry = None

      s.memory = bytearray( 2**20 )

      @s.update
      def up_process():

        if s.entry is not None and s.mem.resp.rdy():

          # Dequeue memory request message

          req     = s.entry
          s.entry = None

          len_ = int(req.len)
          if not len_: len_ = req_class.data_nbits >> 3

          if   req.type_ == MemMsgType.READ:
            resp = resp_class( req.type_, req.opaque, 0, req.len,
                               s.read( req.addr, len_ ) )

          elif req.type_ == MemMsgType.WRITE:
            s.write( req.addr, len_, req.data )
            resp = resp_class( req.type_, req.opaque, 0, 0, 0 )

          s.mem.resp( resp )

      s.add_constraints( U(up_process) < M(s.mem.req) ) # definite pipe behavior

    def line_trace( s ):
      return s.mem.line_trace()

  class TestHarness( Component ):
    def construct( s ):
      s.master = [ SomeMasterFL( i*0x222, i ) for i in range(2) ]
      s.minion = [ SomeMinionCL( *mk_mem_msg(8,32,32) )( mem = s.master[i].mem )
                    for i in range(2) ]

    def line_trace( s ):
      return "|".join( [ x.line_trace() for x in s.master ] ) + " >>> " + \
             "|".join( [ x.line_trace() for x in s.minion ] )

    def done( s ):
      return all( [ x.done() for x in s.master ] )

  th = TestHarness()

  # Create a simulator

  th.apply( SimulationPass() )

  # Run simulation

  print("")
  ncycles = 0
  while not th.done() and ncycles < 1000:
    th.tick()
    ncycles += 1
    trace = th.line_trace()
    print("{:3}: {}".format( ncycles, trace ))

  # Check timeout

  assert ncycles < 1000

  th.tick()
  th.tick()
  th.tick()



#=========================================================================
# TranslationImport_adhoc_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 5, 2019
"""Test ad-hoc components with SystemVerilog translation and import."""

import pytest

from pymtl3.passes.rtlir.util.test_utility import get_parameter
from pymtl3.stdlib.test import TestVectorSimulator

from .. import TranslationImportPass
from ..translation.behavioral.test.SVBehavioralTranslatorL1_test import (
    test_sverilog_behavioral_L1,
)
from ..translation.behavioral.test.SVBehavioralTranslatorL2_test import (
    test_sverilog_behavioral_L2,
)
from ..translation.behavioral.test.SVBehavioralTranslatorL3_test import (
    test_sverilog_behavioral_L3,
)
from ..translation.behavioral.test.SVBehavioralTranslatorL4_test import (
    test_sverilog_behavioral_L4,
)
from ..translation.behavioral.test.SVBehavioralTranslatorL5_test import (
    test_sverilog_behavioral_L5,
)
from ..translation.structural.test.SVStructuralTranslatorL1_test import (
    test_sverilog_structural_L1,
)
from ..translation.structural.test.SVStructuralTranslatorL2_test import (
    test_sverilog_structural_L2,
)
from ..translation.structural.test.SVStructuralTranslatorL3_test import (
    test_sverilog_structural_L3,
)
from ..translation.structural.test.SVStructuralTranslatorL4_test import (
    test_sverilog_structural_L4,
)


def run_test( case ):
  try:
    _m = case.DUT()
    _m.elaborate()
    _m.sverilog_translate_import = True
    m = TranslationImportPass()( _m )
    sim = TestVectorSimulator( m, case.TEST_VECTOR, case.TV_IN, case.TV_OUT )
    sim.run_test()
  finally:
    try:
      m.finalize()
    except UnboundLocalError:
      # This test fails due to translation errors
      pass

@pytest.mark.parametrize(
  'case', get_parameter('case', test_sverilog_behavioral_L1) + \
          get_parameter('case', test_sverilog_behavioral_L2) + \
          get_parameter('case', test_sverilog_behavioral_L3) + \
          get_parameter('case', test_sverilog_behavioral_L4) + \
          get_parameter('case', test_sverilog_behavioral_L5) + \
          get_parameter('case', test_sverilog_structural_L1) + \
          get_parameter('case', test_sverilog_structural_L2) + \
          get_parameter('case', test_sverilog_structural_L3) + \
          get_parameter('case', test_sverilog_structural_L4)
)
def test_sverilog_translation_import_adhoc( case ):
  run_test( case )



#=========================================================================
# BehavioralTranslatorL0.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 22, 2019
"""Provide L0 behavioral translator."""

from ..BaseRTLIRTranslator import BaseRTLIRTranslator, TranslatorMetadata


class BehavioralTranslatorL0( BaseRTLIRTranslator ):
  def __init__( s, top ):
    super().__init__( top )

  def clear( s, tr_top ):
    super().clear( tr_top )
    s.behavioral = TranslatorMetadata()

  def translate_behavioral( s, m ):
    pass



#=========================================================================
# TestRTLIRTranslator.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 23, 2019
"""Provide an RTLIR translator that fits testing purposes."""

from ..behavioral.test.TestBehavioralTranslator import TestBehavioralTranslator
from ..RTLIRTranslator import RTLIRTranslator
from ..structural.test.TestStructuralTranslator import TestStructuralTranslator


def make_indent( src, nindent ):
  """Add nindent indention to every line in src."""
  indent = '  '
  for idx, s in enumerate( src ):
    src[ idx ] = nindent * indent + s

def get_pretty( namespace, attr, newline=True ):
  ret = getattr(namespace, attr, "")
  if newline and (ret and ret[-1] != '\n'):
    ret += "\n"
  return ret

class TestRTLIRTranslator( RTLIRTranslator,
    TestStructuralTranslator, TestBehavioralTranslator ):

  def rtlir_tr_src_layout( s, hierarchy ):
    # struct definitions
    struct_defs = \
      "".join( f"struct {x[1]}\n" for x in hierarchy.decl_type_struct )
    component_src = hierarchy.component_src
    return struct_defs + component_src

  def rtlir_tr_components( s, components ):
    return "\n".join(components)

  def rtlir_tr_component( s, behavioral, structural ):
    template = \
"""\
component {component_name}
(
{ports});
{body}
endcomponent
"""
    ports_template = "{port_decls}{ifc_decls}"
    component_name = structural.component_unique_name

    port_decls = get_pretty(structural, "decl_ports", False)
    ifc_decls = get_pretty(structural, "decl_ifcs", False)
    ports = ports_template.format( **locals() )

    const_decls = get_pretty(structural, "decl_consts")
    fvar_decls = get_pretty(behavioral, "decl_freevars")
    wire_decls = get_pretty(structural, "decl_wires")
    tmpvar_decls = get_pretty(behavioral, "decl_tmpvars")
    subcomp_decls = get_pretty(structural, "decl_subcomps")
    upblk_srcs = get_pretty(behavioral, "upblk_srcs")
    connections = get_pretty(structural, "connections")
    body = const_decls + fvar_decls + wire_decls + subcomp_decls \
         + tmpvar_decls + upblk_srcs + connections

    return template.format( **locals() )



#=========================================================================
# SVImportSignalGen.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 1, 2019
"""Test the SystemVerilog signal generation of imported component."""

from pymtl3.passes.backends.sverilog.import_.test.SVImportSignalGen_test import (
    test_interface,
    test_interface_array,
    test_interface_parameter,
    test_interface_parameter_long_vector,
    test_nested_interface,
    test_nested_interface_port_array,
    test_nested_struct,
    test_packed_array_port_array,
    test_port_2d_array,
    test_port_array,
    test_port_single,
    test_struct_port_array,
    test_struct_port_single,
)
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..ImportPass import ImportPass


def local_do_test( m ):
  m.elaborate()
  rtype = rt.get_component_ifc_rtlir( m )
  ipass = ImportPass()
  symbols, decls, conns = ipass.gen_signal_decl_py( rtype )
  assert conns == m._ref_conns_yosys



"""
==========================================================================
 connect_pairs.py
==========================================================================
Advanced connect functions for different patterns

 Author: Shunning
   Date: July 13, 2019
"""

from pymtl3 import *


def connect_pairs( *args ):

  if len(args) & 1 != 0:
     raise InvalidConnectionError( "Odd number ({}) of objects provided.".format( len(args) ) )

  for i in range(len(args)>>1) :
    try:
      connect( args[ i<<1 ], args[ (i<<1)+1 ] )
    except Exception as e:
      args = list(e.args[::])
      args[0] = "[connect_pair] connecting {}-th argument to {}-th argument\n" \
                .format( i<<1, (i<<1)+1 ) + \
                "--------------------\n" + args[0]
      e.args = tuple(args)
      raise



#=========================================================================
# BehavioralTranslatorL3.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 22, 2019
"""Provide L3 behavioral translator."""

from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRGenL3Pass import (
    BehavioralRTLIRGenL3Pass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRTypeCheckL3Pass import (
    BehavioralRTLIRTypeCheckL3Pass,
)

from .BehavioralTranslatorL2 import BehavioralTranslatorL2


class BehavioralTranslatorL3( BehavioralTranslatorL2 ):
  def __init__( s, top ):
    super().__init__( top )

  def clear( s, tr_top ):
    super().clear( tr_top )

  #-----------------------------------------------------------------------
  # _gen_behavioral_trans_metadata
  #-----------------------------------------------------------------------

  # Override
  def _gen_behavioral_trans_metadata( s, m ):
    m.apply( BehavioralRTLIRGenL3Pass() )
    m.apply( BehavioralRTLIRTypeCheckL3Pass() )
    s.behavioral.rtlir[m] = m._pass_behavioral_rtlir_gen.rtlir_upblks
    s.behavioral.freevars[m] =\
        m._pass_behavioral_rtlir_type_check.rtlir_freevars
    s.behavioral.tmpvars[m] =\
        m._pass_behavioral_rtlir_type_check.rtlir_tmpvars



#=========================================================================
# SVTranslator.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 15, 2019
"""Provide SystemVerilog translator."""

from pymtl3.passes.backends.generic import RTLIRTranslator
from pymtl3.passes.backends.sverilog.util.utility import sverilog_reserved

from .behavioral import SVBehavioralTranslator as SV_BTranslator
from .structural import SVStructuralTranslator as SV_STranslator


def mk_SVTranslator( _RTLIRTranslator, _STranslator, _BTranslator ):

  class _SVTranslator( _RTLIRTranslator, _STranslator, _BTranslator ):

    def get_pretty( s, namespace, attr, newline=True ):
      ret = getattr(namespace, attr, "")
      if newline and (ret and ret[-1] != '\n'):
        ret += "\n"
      return ret

    def is_sverilog_reserved( s, name ):
      return name in sverilog_reserved

    def set_header( s ):
      s.header = \
"""\
//-------------------------------------------------------------------------
// {name}.sv
//-------------------------------------------------------------------------
// This file is generated by PyMTL SystemVerilog translation pass.

"""

    def rtlir_tr_src_layout( s, hierarchy ):
      s.set_header()
      name = s._top_module_full_name
      ret = s.header.format( **locals() )

      # Add struct definitions
      for struct_dtype, tplt in hierarchy.decl_type_struct:
        template = \
"""\
// PyMTL BitStruct {dtype_name} Definition
// At {file_info}
{struct_def}\
"""
        dtype_name = struct_dtype.get_name()
        file_info = struct_dtype.get_file_info()
        struct_def = tplt['def'] + '\n'
        ret += template.format( **locals() )

      # Add component sources
      ret += hierarchy.component_src
      return ret

    def rtlir_tr_components( s, components ):
      return "\n\n".join( components )

    def rtlir_tr_component( s, behavioral, structural ):

      template =\
"""\
// PyMTL Component {component_name} Definition
// {optional_full_name}At {file_info}
module {module_name}
(
{ports});
{body}
endmodule
"""
      component_name = structural.component_name
      file_info = structural.component_file_info
      ports_template = "{port_decls}{ifc_decls}"
      full_name = structural.component_full_name
      module_name = structural.component_unique_name

      if full_name != module_name:
        optional_full_name = f"Full name: {full_name}\n// "
      else:
        optional_full_name = ""

      port_decls = s.get_pretty(structural, 'decl_ports', False)
      ifc_decls = s.get_pretty(structural, 'decl_ifcs', False)
      if port_decls or ifc_decls:
        if port_decls and ifc_decls:
          port_decls += ',\n'
        ifc_decls += '\n'
      ports = ports_template.format(**locals())

      const_decls = s.get_pretty(structural, "decl_consts")
      fvar_decls = s.get_pretty(behavioral, "decl_freevars")
      wire_decls = s.get_pretty(structural, "decl_wires")
      tmpvar_decls = s.get_pretty(behavioral, "decl_tmpvars")
      subcomp_decls = s.get_pretty(structural, "decl_subcomps")
      upblk_decls = s.get_pretty(behavioral, "upblk_decls")
      body = const_decls + fvar_decls + wire_decls + subcomp_decls \
           + tmpvar_decls + upblk_decls
      connections = s.get_pretty(structural, "connections")
      if (body and connections) or (not body and connections):
        connections = '\n' + connections
      body += connections

      s._top_module_name = structural.component_name
      s._top_module_full_name = module_name
      return template.format( **locals() )

  return _SVTranslator

SVTranslator = mk_SVTranslator( RTLIRTranslator, SV_STranslator, SV_BTranslator )



"""
========================================================================
helpers.py
========================================================================
Useful helper functions of Bits operations. Ideally these functions
should be implemented in RPython.

Author : Shunning Jiang
Date   : Nov 3, 2017
"""
import functools
import math
import operator

from .bits_import import *
from .bitstructs import is_bitstruct_class, is_bitstruct_inst


def concat( *args ):
  nbits = sum( [ x.nbits for x in args ] )
  concat_bits = Bits( nbits, 0 )

  begin = 0
  for bits in reversed( args ):
    concat_bits[ begin : begin+bits.nbits ] = bits
    begin += bits.nbits

  return concat_bits

def zext( value, new_width ):
  assert new_width > value.nbits
  return Bits( new_width, value )

def clog2( N ):
  assert N > 0
  return int( math.ceil( math.log( N, 2 ) ) )

def sext( value, new_width ):
  assert new_width > value.nbits
  return Bits( new_width, value.int() )

def reduce_and( value ):
  try:
    return b1( int(value) == (1 << value.nbits) - 1 )
  except AttributeError:
    raise TypeError("Cannot call reduce_and on int")

def reduce_or( value ):
  try:
    return b1( int(value) != 0 )
  except AttributeError:
    raise TypeError("Cannot call reduce_or on int")

def reduce_xor( value ):
  try:
    pop_count = 0
    value = int(value)
    while value != 0:
      pop_count += value & 1
      value >>= 1
    return b1( pop_count & 1 )

  except AttributeError:
    raise TypeError("Cannot call reduce_xor on int")

def get_nbits( Type ):
  if isinstance( Type, list ):
    return sum(get_nbits(v) for v in Type )

  assert isinstance( Type, type )
  if issubclass( Type, Bits ):
    return Type.nbits

  assert is_bitstruct_class( Type ), f"{Type} is not a valid PyMTL data type!"
  return sum(get_nbits(v) for v in Type.__bitstruct_fields__.values())

def to_bits( obj ):
  if isinstance( obj, list ):
    return concat( *[ to_bits(x) for x in obj ] )

  # BitsN
  if isinstance( obj, Bits ):
    return obj

  assert not isinstance( obj, int ), f"{obj} is an integer, not Bits or Bitstruct."
  # BitStruct
  assert is_bitstruct_inst( obj ), f"{obj} is not a valid PyMTL Bitstruct!"
  return concat( *[ to_bits(getattr(obj, v)) for v in obj.__bitstruct_fields__.keys() ] )

def get_bitstruct_inst_all_classes( obj ):
  # list: put all types together
  if isinstance( obj, list ):
    return functools.reduce( operator.or_, [ get_bitstruct_inst_all_classes(x) for x in obj ] )
  ret = { obj.__class__ }
  # BitsN or int
  if isinstance( obj, (Bits, int) ):
    return ret
  # BitStruct
  assert is_bitstruct_inst( obj ), f"{obj} is not a valid PyMTL Bitstruct!"
  return ret | functools.reduce( operator.or_, [ get_bitstruct_inst_all_classes(getattr(obj, v))
                                                for v in obj.__bitstruct_fields__.keys() ] )



#=========================================================================
# StructuralRTLIRGenL4Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Apr 3, 2019
"""Provide L4 structural RTLIR generation pass."""

from pymtl3.passes.BasePass import PassMetadata

from .StructuralRTLIRGenL3Pass import StructuralRTLIRGenL3Pass
from .StructuralRTLIRSignalExpr import gen_signal_expr


class StructuralRTLIRGenL4Pass( StructuralRTLIRGenL3Pass ):
  # Override
  def __call__( s, tr_top ):
    s.gen_metadata( tr_top )
    super().__call__( tr_top )

  def gen_metadata( s, m ):
    if not hasattr( m, '_pass_structural_rtlir_gen' ):
      m._pass_structural_rtlir_gen = PassMetadata()
    for child in m.get_child_components():
      s.gen_metadata( child )

  # Override
  def gen_rtlir_types( s, m ):
    super().gen_rtlir_types( m )
    for child in m.get_child_components():
      s.gen_rtlir_types( child )

  # Override
  def gen_constants( s, m ):
    super().gen_constants( m )
    for child in m.get_child_components():
      s.gen_constants( child )

  # Override
  def sort_connections( s, m ):
    """Sort connections by the order `connect` is called.

    At L4 we need to recursively generate connections for every component.
    """
    super().sort_connections( m )

    for child in m.get_child_components():
      s.sort_connections( child )



"""
========================================================================
Bits.py
========================================================================
Pure-Python implementation of fixed-bitwidth data type.

Author : Shunning Jiang
Date   : Oct 31, 2017
"""


class Bits:
  __slots__ = ( "nbits", "value" )

  def __init__( self, nbits=32, value=0 ):
    self.nbits = nbits
    self.value = int(value) & ((1 << nbits) - 1)

  def __ilshift__( self, x ):
    try:
      assert x.nbits == self.nbits, "Bitwidth mismatch during <<="
    except AttributeError:
      raise TypeError(f"Assign {type(x)} to Bits")
    self._next = x.value
    return self

  def _flip( self ):
    self.value = self._next

  def clone( self ):
    return Bits( self.nbits, self.value )

  def __deepcopy__( self, memo ):
    return Bits( self.nbits, self.value )

  def __call__( self ):
    return Bits( self.nbits )

  # Arithmetics
  def __getitem__( self, idx ):
    sv = int(self.value)

    if isinstance( idx, slice ):
      start, stop = int(idx.start), int(idx.stop)
      assert not idx.step and start < stop and start >= 0 and stop <= self.nbits, \
            "Invalid access: [{}:{}] in a Bits{} instance".format( start, stop, self.nbits )
      return Bits( stop-start, (sv & ((1 << stop) - 1)) >> start )

    i = int(idx)
    assert 0 <= i < self.nbits
    return Bits( 1, (sv >> i) & 1 )

  def __setitem__( self, idx, v ):
    sv = int(self.value)

    if isinstance( idx, slice ):
      start, stop = int(idx.start), int(idx.stop)
      assert not idx.step and start < stop and start >= 0 and stop <= self.nbits, \
            "Invalid access: [{}:{}] in a Bits{} instance".format( start, stop, self.nbits )

      self.value = (sv & (~((1 << stop) - (1 << start)))) | \
                   ((int(v) & ((1 << (stop - start)) - 1)) << start)
      return

    i = int(idx)
    assert 0 <= i < self.nbits
    self.value = (sv & ~(1 << i)) | ((int(v) & 1) << i)

  def __add__( self, other ):
    try:    return Bits( max(self.nbits, other.nbits), int(self.value) + int(other) )
    except: return Bits( self.nbits, int(self.value) + int(other) )

  def __radd__( self, other ):
    return self.__add__( other )

  def __sub__( self, other ):
    try:    return Bits( max(self.nbits, other.nbits), int(self.value) - int(other) )
    except: return Bits( self.nbits, int(self.value) - int(other) )

  def __rsub__( self, other ):
    return Bits( self.nbits, other ) - self

  def __mul__( self, other ):
    try:    return Bits( max(self.nbits, other.nbits), int(self.value) * int(other) )
    except: return Bits( self.nbits, int(self.value) * int(other) )

  def __rmul__( self, other ):
    return self.__mul__( other )

  def __and__( self, other ):
    try:    return Bits( max(self.nbits, other.nbits), int(self.value) & int(other) )
    except: return Bits( self.nbits, int(self.value) & int(other) )

  def __rand__( self, other ):
    return self.__and__( other )

  def __or__( self, other ):
    try:    return Bits( max(self.nbits, other.nbits), int(self.value) | int(other) )
    except: return Bits( self.nbits, int(self.value) | int(other) )

  def __ror__( self, other ):
    return self.__or__( other )

  def __xor__( self, other ):
    try:    return Bits( max(self.nbits, other.nbits), int(self.value) ^ int(other) )
    except: return Bits( self.nbits, int(self.value) ^ int(other) )

  def __rxor__( self, other ):
    return self.__xor__( other )

  def __div__( self, other ):
    try:    return Bits( max(self.nbits, other.nbits), int(self.value) / int(other) )
    except: return Bits( self.nbits, int(self.value) / int(other) )

  def __floordiv__( self, other ):
    try:    return Bits( max(self.nbits, other.nbits), int(self.value) / int(other) )
    except: return Bits( self.nbits, int(self.value) / int(other) )

  def __mod__( self, other ):
    try:    return Bits( max(self.nbits, other.nbits), int(self.value) % int(other) )
    except: return Bits( self.nbits, int(self.value) % int(other) )

  def __invert__( self ):
    return Bits( self.nbits, ~int(self.value) )

  def __lshift__( self, other ):
    nb = int(self.nbits)
    # TODO this doesn't work perfectly. We need a really smart
    # optimization that avoids the guard totally
    if int(other) >= nb: return Bits( nb )
    return Bits( nb, int(self.value) << int(other) )

  def __rshift__( self, other ):
    return Bits( self.nbits, int(self.value) >> int(other) )

  def __eq__( self, other ):
    try:
      other = int(other)
    except ValueError:
      return False
    return Bits( 1, int(self.value) == other )

  def __hash__( self ):
    return hash((self.nbits, self.value))

  def __ne__( self, other ):
    try:
      other = int(other)
    except ValueError:
      return True
    return Bits( 1, int(self.value) != other )

  def __lt__( self, other ):
    return Bits( 1, int(self.value) < int(other) )

  def __le__( self, other ):
    return Bits( 1, int(self.value) <= int(other) )

  def __gt__( self, other ):
    return Bits( 1, int(self.value) > int(other) )

  def __ge__( self, other ):
    return Bits( 1, int(self.value) >= int(other) )

  def __bool__( self ):
    return int(self.value) != 0

  def __int__( self ):
    return int(self.value)

  def int( self ):
    if self.value >> (self.nbits - 1):
      return -int(~self + 1)
    return int(self.value)

  def uint( self ):
    return int(self.value)

  def __index__( self ):
    return int(self.value)

  # Print

  def __repr__(self):
    return "Bits{}({})".format( self.nbits, int(self.value) )

  def __str__(self):
    str = "{:x}".format(int(self.value)).zfill(((self.nbits-1)>>2)+1)
    return str

  def __oct__( self ):
    # print("DEPRECATED: Please use .oct()!")
    return self.oct()

  def __hex__( self ):
    # print("DEPRECATED: Please use .hex()!")
    return self.hex()

  def bin(self):
    str = "{:b}".format(int(self.value)).zfill(self.nbits)
    return "0b"+str

  def oct( self ):
    str = "{:o}".format(int(self.value)).zfill(((self.nbits-1)>>1)+1)
    return "0o"+str

  def hex( self ):
    str = "{:x}".format(int(self.value)).zfill(((self.nbits-1)>>2)+1)
    return "0x"+str



#=========================================================================
# YosysTranslator.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 8, 2019
"""Provide yosys-compatible SystemVerilog translator."""

from pymtl3.passes.backends.generic import RTLIRTranslator
from pymtl3.passes.backends.sverilog.translation.SVTranslator import mk_SVTranslator

from .behavioral import YosysBehavioralTranslator as Yosys_BTranslator
from .structural import YosysStructuralTranslator as Yosys_STranslator

SVTranslator = mk_SVTranslator( RTLIRTranslator, Yosys_STranslator, Yosys_BTranslator )

class YosysTranslator( SVTranslator ):

  def set_header( s ):
      s.header = \
"""\
//-------------------------------------------------------------------------
// {name}.sv
//-------------------------------------------------------------------------
// This file is generated by PyMTL yosys-SystemVerilog translation pass.

"""

  def rtlir_tr_src_layout( s, hierarchy ):
    s.set_header()
    name = s._top_module_full_name
    ret = s.header.format( **locals() )

    # Add component sources
    ret += hierarchy.component_src
    return ret

  def rtlir_tr_component( s, behavioral, structural ):

    template =\
"""\
// PyMTL Component {component_name} Definition
// {optional_full_name}At {file_info}
module {module_name}
(
{ports});
{body}
endmodule
"""
    component_name = structural.component_name
    file_info = structural.component_file_info
    ports_template = "{port_decls}{ifc_decls}"
    full_name = structural.component_full_name
    module_name = structural.component_unique_name

    if full_name != module_name:
      optional_full_name = f"Full name: {full_name}\n// "
    else:
      optional_full_name = ""

    port_dct = structural.decl_ports
    structural.p_port_decls = port_dct["port_decls"]
    structural.p_wire_decls = port_dct["wire_decls"]
    structural.p_connections = port_dct["connections"]

    ifc_dct = structural.decl_ifcs
    structural.i_port_decls = ifc_dct["port_decls"]
    structural.i_wire_decls = ifc_dct["wire_decls"]
    structural.i_connections = ifc_dct["connections"]

    subcomp_dct = structural.decl_subcomps
    structural.c_port_decls = subcomp_dct["port_decls"]
    structural.c_wire_decls = subcomp_dct["wire_decls"]
    structural.c_connections = subcomp_dct["connections"]

    # Assemble ports and interfaces
    port_decls = s.get_pretty(structural, 'p_port_decls', False)
    ifc_decls = s.get_pretty(structural, 'i_port_decls', False)
    if port_decls or ifc_decls:
      if port_decls and ifc_decls:
        port_decls += ',\n'
      ifc_decls += '\n'
    ports = ports_template.format(**locals())

    # Assemble body of module definition

    # Begin with port_wires
    p_port_wires = s.get_pretty(structural, "p_wire_decls", False)
    i_port_wires = s.get_pretty(structural, "i_wire_decls", False)

    if p_port_wires or i_port_wires:
      if p_port_wires and i_port_wires:
        p_port_wires += "\n"
      i_port_wires += "\n"
    port_wires = p_port_wires + i_port_wires
    if port_wires:
      port_wires = \
          "  // Struct/Array ports in the form of wires\n" \
          + port_wires

    body = port_wires

    # Add wire declarations
    wire_decls = s.get_pretty(structural, "decl_wires")
    if wire_decls:
      wire_decls = "  // Wire declarations\n" + wire_decls
    if body and wire_decls:
      wire_decls = "\n" + wire_decls
    body += wire_decls

    # Add wires for sub-component struct/array ports
    subcomp_wires = s.get_pretty(structural, "c_wire_decls")
    if subcomp_wires:
      subcomp_wires = \
          "  // Struct/Array ports of sub-components in the form of wires\n" \
          + subcomp_wires
    if body and subcomp_wires:
      subcomp_wires = "\n" + subcomp_wires
    body += subcomp_wires

    # Add sub-component declarations
    subcomp_ports = s.get_pretty(structural, "c_port_decls")
    if subcomp_ports:
      subcomp_ports = "  // Sub-component declarations\n" + subcomp_ports
    if body and subcomp_ports:
      subcomp_ports = "\n" + subcomp_ports
    body += subcomp_ports

    # Add connections between wires and sub-component struct/array ports
    subcomp_conns = s.get_pretty(structural, "c_connections")
    if subcomp_conns:
      subcomp_conns = "  // Connect struct/array ports and their wire forms\n" \
                      + subcomp_conns
    if body and subcomp_conns:
      subcomp_conns = "\n" + subcomp_conns
    body += subcomp_conns

    # Add temporary wire definitions
    tmpvar_decls = s.get_pretty(behavioral, "decl_tmpvars")
    if tmpvar_decls:
      tmpvar_decls = "  // Temporary wire definitions\n" + tmpvar_decls
    if body and tmpvar_decls:
      tmpvar_decls = "\n" + tmpvar_decls
    body += tmpvar_decls

    # Add procedural blocks
    upblk_decls = s.get_pretty(behavioral, "upblk_decls")
    body += upblk_decls

    # Add connections
    p_conns = s.get_pretty(structural, "p_connections", False)
    i_conns = s.get_pretty(structural, "i_connections", False)

    if p_conns or i_conns:
      if p_conns and i_conns:
        p_conns += "\n"
      i_conns += "\n"
    port_connections = p_conns + i_conns
    connections = port_connections \
                + s.get_pretty(structural, "connections")
    if connections:
      connections = "  // Connections\n" + connections
    if (body and connections) or (not body and connections):
      connections = '\n' + connections
    body += connections

    s._top_module_name = structural.component_name
    s._top_module_full_name = module_name

    # Fill in the template and return
    return template.format( **locals() )



"""
==========================================================================
helpers_test.py
==========================================================================
Test cases for helper functions

Author : Shunning Jiang
  Date : Nov 30, 2019
"""

from pymtl3.datatypes import *
from pymtl3.datatypes.helpers import get_bitstruct_inst_all_classes


def test_concat():
  x = concat( Bits128(0x1234567890abcdef1234567890abcdef),
              Bits252(0xffffffffff22222222222222222444441234567890abcdef1234567890abcdef) )
  assert x.nbits == 380
  assert x == mk_bits(380)(0x1234567890abcdef1234567890abcdeffffffffff22222222222222222444441234567890abcdef1234567890abcdef)

def test_zext():
  assert zext( Bits8(0xe), 24 ) == Bits24(0xe)

def test_sext():
  assert zext( Bits8(0xe), 24 ) == Bits24(0xe)

def test_clog2():
  assert clog2(7) == 3
  assert clog2(8) == 3
  assert clog2(9) == 4

def test_reduce_and():
  for i in range(255):
    assert not reduce_and( b8(i) )
  assert reduce_and( b8(255) )
  assert reduce_and( b512((2**512)-1) )

def test_reduce_or():
  for i in range(1, 256):
    assert reduce_or(i)
  assert not reduce_or( b8(0) )
  assert reduce_or( b512((2**512)-1) )

def test_reduce_xor():
  assert reduce_xor( b8(0b10101011) ) == 1
  assert reduce_xor( b8(0b10101010) ) == 0

def test_get_nbits():

  @bitstruct
  class SomeMsg:
    c: [ Bits3, Bits3 ]
    d: [ Bits5, Bits5 ]

  assert get_nbits( SomeMsg ) == 16

def test_to_bits():

  @bitstruct
  class SomeMsg:
    c: [ Bits4, Bits4 ]
    d: [ Bits8, Bits8 ]

  assert to_bits( SomeMsg([b4(0x2),b4(0x3)],[b8(0x45), b8(0x67)]) ) == b24(0x234567)

def test_get_bitstruct_inst_all_classes():

  @bitstruct
  class SomeMsg1:
    a: [ Bits4, Bits4 ]
    b: Bits8

  @bitstruct
  class SomeMsg2:
    c: [ SomeMsg1, SomeMsg1 ]
    d: [ Bits6, Bits6 ]

  a = SomeMsg2()
  print()
  print(get_bitstruct_inst_all_classes( a ))
  print({Bits4, Bits8, SomeMsg1, Bits6, SomeMsg2})
  assert get_bitstruct_inst_all_classes( a ) == {Bits4, Bits8, SomeMsg1, Bits6, SomeMsg2}



"""
========================================================================
ValRdyIfc
========================================================================
RTL val/rdy interface.

Author : Shunning Jiang
  Date : Apr 5, 2019
"""
from pymtl3 import *

from .ifcs_utils import valrdy_to_str


class InValRdyIfc( Interface ):

  def construct( s, Type ):

    s.msg = InPort( Type )
    s.val = InPort( int if Type is int else Bits1 )
    s.rdy = OutPort( int if Type is int else Bits1 )

  def line_trace( s ):
    return valrdy_to_str( s.msg, s.val, s.rdy )

class OutValRdyIfc( Interface ):

  def construct( s, Type ):

    s.msg = OutPort( Type )
    s.val = OutPort( int if Type is int else Bits1 )
    s.rdy = InPort( int if Type is int else Bits1 )

  def line_trace( s ):
    return valrdy_to_str( s.msg, s.val, s.rdy )



#=========================================================================
# SVBehavioralTranslatorL2_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 28, 2019
"""Test the SystemVerilog translator implementation."""

import pytest

from pymtl3.passes.backends.sverilog.util.utility import sverilog_reserved
from pymtl3.passes.rtlir import BehavioralRTLIRGenPass, BehavioralRTLIRTypeCheckPass

from ....testcases import (
    CaseElifBranchComp,
    CaseFixedSizeSliceComp,
    CaseForRangeLowerUpperStepPassThroughComp,
    CaseIfBasicComp,
    CaseIfBoolOpInForStmtComp,
    CaseIfDanglingElseInnerComp,
    CaseIfDanglingElseOutterComp,
    CaseIfExpInForStmtComp,
    CaseIfExpUnaryOpInForStmtComp,
    CaseIfTmpVarInForStmtComp,
    CaseLambdaConnectComp,
    CaseLambdaConnectWithListComp,
    CaseNestedIfComp,
    CaseReducesInx3OutComp,
)
from ..SVBehavioralTranslatorL2 import BehavioralRTLIRToSVVisitorL2


def run_test( case, m ):
  m.elaborate()
  m.apply( BehavioralRTLIRGenPass() )
  m.apply( BehavioralRTLIRTypeCheckPass() )

  visitor = BehavioralRTLIRToSVVisitorL2(lambda x: x in sverilog_reserved)
  upblks = m._pass_behavioral_rtlir_gen.rtlir_upblks
  m_all_upblks = m.get_update_blocks()
  assert len(m_all_upblks) == 1

  for blk in m_all_upblks:
    upblk_src = visitor.enter( blk, upblks[blk] )
    upblk_src = "\n".join( upblk_src )
    assert upblk_src + '\n' == case.REF_UPBLK

@pytest.mark.parametrize(
    'case', [
      CaseReducesInx3OutComp,
      CaseIfBasicComp,
      CaseIfDanglingElseInnerComp,
      CaseIfDanglingElseOutterComp,
      CaseElifBranchComp,
      CaseNestedIfComp,
      CaseForRangeLowerUpperStepPassThroughComp,
      CaseIfExpInForStmtComp,
      CaseIfBoolOpInForStmtComp,
      CaseIfTmpVarInForStmtComp,
      CaseFixedSizeSliceComp,
      CaseLambdaConnectComp,
      CaseLambdaConnectWithListComp,
    ]
)
def test_sverilog_behavioral_L2( case ):
  run_test( case, case.DUT() )

@pytest.mark.xfail(run=False, reason="TODO: resolving BitStructs according to name AND fields")
def test_struct_uniqueness():
  class A:
    @bitstruct
    class ST:
      a_foo: Bits16
      a_bar: Bits32

  class B:
    @bitstruct
    class ST:
      b_foo: Bits16
      b_bar: Bits32

  @bitstruct
  class COMB:
    fst: A.ST
    snd: B.ST

  class Top( Component ):
    def construct( s ):
      s.out = OutPort( COMB )
      connect( s.out, COMB(A.ST(1, 2), B.ST(3, 4)) )
  a = Top()
  a.REF_SRC = \
"""
typedef struct packed {
  logic [15:0] foo;
  logic [31:0] bar;
} ST;

typedef struct packed {
  ST fst;
  ST snd;
} COMB;

module Top
(
  input logic [0:0] clk,
  output COMB out,
  input logic [0:0] reset
);

  assign out = { { 16'd1, 32'd2 }, { 16'd3, 32'd4 } };

endmodule
"""
  run_test( a, Top() )
